@@ -1,13 +1,9 @@
 /*
-  $Header: /cvs/src/chrony/client.c,v 1.68 2003/09/22 21:22:30 richard Exp $
-
-  =======================================================================
-
   chronyd/chronyc - Programs for keeping computer clocks accurate.
 
  **********************************************************************
  * Copyright (C) Richard P. Curnow  1997-2003
- * Copyright (C) Miroslav Lichvar  2009
+ * Copyright (C) Miroslav Lichvar  2009-2011
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of version 2 of the GNU General Public License as
@@ -30,12 +26,13 @@
   from it whilst running.
   */
 
+#include "config.h"
+
 #include "sysincl.h"
 
 #include "candm.h"
 #include "nameserv.h"
 #include "md5.h"
-#include "version.h"
 #include "getdate.h"
 #include "cmdparse.h"
 #include "pktlength.h"
@@ -150,7 +147,7 @@ open_io(const char *hostname, int port)
   IPAddr ip;
 
   /* Note, this call could block for a while */
-  if (!DNS_Name2IPAddress(hostname, &ip, 0)) {
+  if (DNS_Name2IPAddress(hostname, &ip) != DNS_Success) {
     fprintf(stderr, "Could not get IP address for %s\n", hostname);
     exit(1);
   }
@@ -330,7 +327,7 @@ read_address_integer(char *line, IPAddr *address, int *value)
     fprintf(stderr, "Invalid syntax for address value\n");
     ok = 0;
   } else {
-    if (!DNS_Name2IPAddress(hostname, address, 0)) {
+    if (DNS_Name2IPAddress(hostname, address) != DNS_Success) {
       fprintf(stderr, "Could not get address for hostname\n");
       ok = 0;
     } else {
@@ -355,7 +352,7 @@ read_address_double(char *line, IPAddr *address, double *value)
     fprintf(stderr, "Invalid syntax for address value\n");
     ok = 0;
   } else {
-    if (!DNS_Name2IPAddress(hostname, address, 0)) {
+    if (DNS_Name2IPAddress(hostname, address) != DNS_Success) {
       fprintf(stderr, "Could not get address for hostname\n");
       ok = 0;
     } else {
@@ -436,6 +433,28 @@ process_cmd_maxdelay(CMD_Request *msg, char *line)
 
 /* ================================================== */
 
+static int
+process_cmd_maxdelaydevratio(CMD_Request *msg, char *line)
+{
+  IPAddr address;
+  double max_delay_dev_ratio;
+  int ok;
+  
+  if (read_address_double(line, &address, &max_delay_dev_ratio)) {
+    UTI_IPHostToNetwork(&address, &msg->data.modify_maxdelaydevratio.address);
+    msg->data.modify_maxdelayratio.new_max_delay_ratio = UTI_FloatHostToNetwork(max_delay_dev_ratio);
+    msg->command = htons(REQ_MODIFY_MAXDELAYDEVRATIO);
+    ok = 1;
+  } else {
+    ok = 0;
+  }
+
+  return ok;
+
+}
+
+/* ================================================== */
+
 static int
 process_cmd_maxdelayratio(CMD_Request *msg, char *line)
 {
@@ -458,6 +477,50 @@ process_cmd_maxdelayratio(CMD_Request *msg, char *line)
 
 /* ================================================== */
 
+static int
+process_cmd_minstratum(CMD_Request *msg, char *line)
+{
+  IPAddr address;
+  int min_stratum;
+  int ok;
+  
+  if (read_address_integer(line, &address, &min_stratum)) {
+    UTI_IPHostToNetwork(&address, &msg->data.modify_minstratum.address);
+    msg->data.modify_minstratum.new_min_stratum = htonl(min_stratum);
+    msg->command = htons(REQ_MODIFY_MINSTRATUM);
+    ok = 1;
+  } else {
+    ok = 0;
+  }
+
+  return ok;
+
+}
+
+/* ================================================== */
+
+static int
+process_cmd_polltarget(CMD_Request *msg, char *line)
+{
+  IPAddr address;
+  int poll_target;
+  int ok;
+  
+  if (read_address_integer(line, &address, &poll_target)) {
+    UTI_IPHostToNetwork(&address, &msg->data.modify_polltarget.address);
+    msg->data.modify_polltarget.new_poll_target = htonl(poll_target);
+    msg->command = htons(REQ_MODIFY_POLLTARGET);
+    ok = 1;
+  } else {
+    ok = 0;
+  }
+
+  return ok;
+
+}
+
+/* ================================================== */
+
 static int
 process_cmd_maxupdateskew(CMD_Request *msg, char *line)
 {
@@ -621,7 +684,7 @@ parse_allow_deny(CMD_Request *msg, char *line)
         if (*q == '\n') *q = 0;
         q++;
       }
-      if (!DNS_Name2IPAddress(p, &ip, 0)) {
+      if (DNS_Name2IPAddress(p, &ip) != DNS_Success) {
         fprintf(stderr, "Could not read address\n");
         return 0;
       } else {
@@ -795,7 +858,7 @@ accheck_getaddr(char *line, IPAddr *addr)
         if (*q == '\n') *q = 0;
         q++;
       }
-      if (!DNS_Name2IPAddress(p, &ip, 0)) {
+      if (DNS_Name2IPAddress(p, &ip) != DNS_Success) {
         return 0;
       } else {
         *addr = ip;
@@ -902,6 +965,28 @@ process_cmd_add_server_or_peer(CMD_Request *msg, char *line)
   status = CPS_ParseNTPSourceAdd(line, &data);
   switch (status) {
     case CPS_Success:
+      /* Don't retry name resolving */
+      if (data.ip_addr.family == IPADDR_UNSPEC) {
+        Free(data.name);
+        fprintf(stderr, "Invalid host/IP address\n");
+        break;
+      }
+
+      if (data.params.min_stratum != SRC_DEFAULT_MINSTRATUM) {
+        fprintf(stderr, "Option minstratum not supported\n");
+        break;
+      }
+
+      if (data.params.poll_target != SRC_DEFAULT_POLLTARGET) {
+        fprintf(stderr, "Option polltarget not supported\n");
+        break;
+      }
+
+      if (data.params.max_delay_dev_ratio != SRC_DEFAULT_MAXDELAYDEVRATIO) {
+        fprintf(stderr, "Option maxdelaydevratio not supported\n");
+        break;
+      }
+
       msg->data.ntp_source.port = htonl((unsigned long) data.port);
       UTI_IPHostToNetwork(&data.ip_addr, &msg->data.ntp_source.ip_addr);
       msg->data.ntp_source.minpoll = htonl(data.params.minpoll);
@@ -912,7 +997,10 @@ process_cmd_add_server_or_peer(CMD_Request *msg, char *line)
       msg->data.ntp_source.max_delay_ratio = UTI_FloatHostToNetwork(data.params.max_delay_ratio);
       msg->data.ntp_source.flags = htonl(
           (data.params.online ? REQ_ADDSRC_ONLINE : 0) |
-          (data.params.auto_offline ? REQ_ADDSRC_AUTOOFFLINE : 0));
+          (data.params.auto_offline ? REQ_ADDSRC_AUTOOFFLINE : 0) |
+          (data.params.iburst ? REQ_ADDSRC_IBURST : 0) |
+          (data.params.sel_option == SRC_SelectPrefer ? REQ_ADDSRC_PREFER : 0) |
+          (data.params.sel_option == SRC_SelectNoselect ? REQ_ADDSRC_NOSELECT : 0));
       result = 1;
 
       break;
@@ -934,6 +1022,9 @@ process_cmd_add_server_or_peer(CMD_Request *msg, char *line)
     case CPS_BadPresend:
       fprintf(stderr, "Unreadable presend value\n");
       break;
+    case CPS_BadMaxdelaydevratio:
+      fprintf(stderr, "Unreadable max delay dev ratio value\n");
+      break;
     case CPS_BadMaxdelayratio:
       fprintf(stderr, "Unreadable max delay ratio value\n");
       break;
@@ -943,6 +1034,12 @@ process_cmd_add_server_or_peer(CMD_Request *msg, char *line)
     case CPS_BadKey:
       fprintf(stderr, "Unreadable key value\n");
       break;
+    case CPS_BadMinstratum:
+      fprintf(stderr, "Unreadable minstratum value\n");
+      break;
+    case CPS_BadPolltarget:
+      fprintf(stderr, "Unreadable polltarget value\n");
+      break;
   }
 
   return result;
@@ -981,16 +1078,15 @@ process_cmd_delete(CMD_Request *msg, char *line)
     fprintf(stderr, "Invalid syntax for address\n");
     ok = 0;
   } else {
-    if (!DNS_Name2IPAddress(hostname, &address, 0)) {
+    if (DNS_Name2IPAddress(hostname, &address) != DNS_Success) {
       fprintf(stderr, "Could not get address for hostname\n");
       ok = 0;
     } else {
+      UTI_IPHostToNetwork(&address, &msg->data.del_source.ip_addr);
       ok = 1;
     }
   }
 
-  UTI_IPHostToNetwork(&address, &msg->data.del_source.ip_addr);
-
   return ok;
   
 }
@@ -1007,7 +1103,6 @@ process_cmd_password(CMD_Request *msg, char *line)
 {
   char *p, *q;
   char *password;
-  struct timezone tz;
   struct timeval now;
 
   p = line;
@@ -1041,7 +1136,7 @@ process_cmd_password(CMD_Request *msg, char *line)
     *p = 0;
   }
     
-  if (gettimeofday(&now, &tz) < 0) {
+  if (gettimeofday(&now, NULL) < 0) {
     printf("500 - Could not read time of day\n");
     return 0;
   } else {
@@ -1123,22 +1218,29 @@ give_help(void)
   printf("manual list : Show previous settime entries\n");
   printf("maxdelay <address> <new-max-delay> : Modify maximum round-trip valid sample delay for source\n");
   printf("maxdelayratio <address> <new-max-ratio> : Modify max round-trip delay ratio for source\n");
-  printf("maxpoll <address> <new-minpoll> : Modify maximum polling interval of source\n");
+  printf("maxdelaydevratio <address> <new-max-ratio> : Modify max round-trip delay dev ratio for source\n");
+  printf("maxpoll <address> <new-maxpoll> : Modify maximum polling interval of source\n");
   printf("maxupdateskew <new-max-skew> : Modify maximum skew for a clock frequency update to be made\n");
   printf("minpoll <address> <new-minpoll> : Modify minimum polling interval of source\n");
+  printf("minstratum <address> <new-min-stratum> : Modify minimum stratum of source\n");
   printf("offline [<mask>/<masked-address>] : Set sources in subnet to offline status\n");
   printf("online [<mask>/<masked-address>] : Set sources in subnet to online status\n");
   printf("password [<new-password>] : Set command authentication password\n");
+  printf("polltarget <address> <new-poll-target> : Modify poll target of source\n");
+  printf("reselect : Reselect synchronisation source\n");
   printf("rtcdata : Print current RTC performance parameters\n");
   printf("settime <date/time (e.g. Nov 21, 1997 16:30:05 or 16:30:05)> : Manually set the daemon time\n");
   printf("sources [-v] : Display information about current sources\n");
   printf("sourcestats [-v] : Display estimation information about current sources\n");
   printf("tracking : Display system time information\n");
   printf("trimrtc : Correct RTC relative to system clock\n");
+  printf("waitsync [max-tries [max-correction [max-skew]]] : Wait until synchronised\n");
   printf("writertc : Save RTC parameters to file\n");
   printf("\n");
   printf("dns -n|+n : Disable/enable resolving IP addresses to hostnames\n");
   printf("dns -4|-6|-46 : Resolve hostnames only to IPv4/IPv6/both addresses\n");
+  printf("timeout <milliseconds> : Set initial response timeout\n");
+  printf("retries <n> : Set maximum number of retries\n");
   printf("exit|quit : Leave the program\n");
   printf("help : Generate this help\n");
   printf("\n");
@@ -1150,8 +1252,8 @@ static unsigned long sequence = 0;
 static unsigned long utoken = 0;
 static unsigned long token = 0;
 
-#define MAX_ATTEMPTS 5
-
+static int max_retries = 2;
+static int initial_timeout = 1000;
 
 /* This is the core protocol module.  Complete particular fields in
    the outgoing packet, send it, wait for a response, handle retries,
@@ -1170,8 +1272,8 @@ submit_request(CMD_Request *request, CMD_Reply *reply, int *reply_auth_ok)
   int read_length;
   int expected_length;
   int command_length;
-  struct timeval timeout;
-  int timeout_seconds;
+  struct timeval tv;
+  int timeout;
   int n_attempts;
   fd_set rdfd, wrfd, exfd;
 
@@ -1185,8 +1287,7 @@ submit_request(CMD_Request *request, CMD_Reply *reply, int *reply_auth_ok)
   request->utoken = htonl(utoken);
   request->token = htonl(token);
 
-
-  timeout_seconds = 2;
+  timeout = initial_timeout;
 
   n_attempts = 0;
 
@@ -1205,6 +1306,8 @@ submit_request(CMD_Request *request, CMD_Reply *reply, int *reply_auth_ok)
     }
 
     command_length = PKL_CommandLength(request);
+    assert(command_length > 0);
+
 #if 0
     printf("Sent command length=%d bytes\n", command_length);
 #endif
@@ -1222,17 +1325,17 @@ submit_request(CMD_Request *request, CMD_Reply *reply, int *reply_auth_ok)
     /* Increment this for next time */
     ++ request->attempt;
       
-    timeout.tv_sec = timeout_seconds;
-    timeout.tv_usec = 0;
+    tv.tv_sec = timeout / 1000;
+    tv.tv_usec = timeout % 1000 * 1000;
+    timeout *= 2;
 
-    timeout_seconds += 1;
     FD_ZERO(&rdfd);
     FD_ZERO(&wrfd);
     FD_ZERO(&exfd);
 
     FD_SET(sock_fd, &rdfd);
 
-    select_status = select(sock_fd + 1, &rdfd, &wrfd, &exfd, &timeout);
+    select_status = select(sock_fd + 1, &rdfd, &wrfd, &exfd, &tv);
 
     if (select_status < 0) {
 #if 0
@@ -1241,7 +1344,7 @@ submit_request(CMD_Request *request, CMD_Reply *reply, int *reply_auth_ok)
     } else if (select_status == 0) {
       /* Timeout must have elapsed, try a resend? */
       n_attempts ++;
-      if (n_attempts == MAX_ATTEMPTS) {
+      if (n_attempts > max_retries) {
         return 0;
       }
 
@@ -1264,7 +1367,7 @@ submit_request(CMD_Request *request, CMD_Reply *reply, int *reply_auth_ok)
            going to a dead port - but only if the daemon machine is
            running Linux (Solaris doesn't return anything) */
         n_attempts++;
-        if (n_attempts == MAX_ATTEMPTS) {
+        if (n_attempts > max_retries) {
           return 0;
         }
       } else {
@@ -1293,7 +1396,7 @@ submit_request(CMD_Request *request, CMD_Reply *reply, int *reply_auth_ok)
         
         if (bad_length || bad_sender || bad_sequence) {
           n_attempts++;
-          if (n_attempts == MAX_ATTEMPTS) {
+          if (n_attempts > max_retries) {
             return 0;
           }
           continue;
@@ -1309,7 +1412,7 @@ submit_request(CMD_Request *request, CMD_Reply *reply, int *reply_auth_ok)
         
         if (bad_header) {
           n_attempts++;
-          if (n_attempts == MAX_ATTEMPTS) {
+          if (n_attempts > max_retries) {
             return 0;
           }
           continue;
@@ -1483,21 +1586,26 @@ print_seconds(unsigned long s)
 static void
 print_nanoseconds(double s)
 {
-  unsigned long ms, ns;
+  s = fabs(s);
 
-  ns = s * 1e9 + 0.5;
-  ms = s * 1e3 + 0.5;
-
-  if (ns <= 9999) {
-    printf("%4ldns", ns);
-  } else if (ns <= 9999499) {
-    printf("%4ldus", (ns + 500) / 1000);
-  } else if (ms <= 9999) {
-    printf("%4ldms", ms);
-  } else if (ms <= 999949) {
-    printf("%3ld.%01lds", (ms + 50) / 1000, ((ms + 50) / 100) % 10);
+  if (s < 9999.5e-9) {
+    printf("%4.0fns", s * 1e9);
+  } else if (s < 9999.5e-6) {
+    printf("%4.0fus", s * 1e6);
+  } else if (s < 9999.5e-3) {
+    printf("%4.0fms", s * 1e3);
+  } else if (s < 999.5) {
+    printf("%5.1fs", s);
+  } else if (s < 99999.5) {
+    printf("%5.0fs", s);
+  } else if (s < 99999.5 * 60) {
+    printf("%5.0fm", s / 60);
+  } else if (s < 99999.5 * 3600) {
+    printf("%5.0fh", s / 3600);
+  } else if (s < 99999.5 * 3600 * 24) {
+    printf("%5.0fd", s / (3600 * 24));
   } else {
-    printf("%5lds", (ms + 500) / 1000);
+    printf("%5.0fy", s / (3600 * 24 * 365));
   }
 }
 
@@ -1506,26 +1614,40 @@ print_nanoseconds(double s)
 static void
 print_signed_nanoseconds(double s)
 {
-  long ms, ns, sign;
+  double x;
 
-  if (s >= 0.0) {
-    ns = s * 1e9 + 0.5;
-    ms = s * 1e3 + 0.5;
-    sign = 1;
+  x = fabs(s);
+
+  if (x < 9999.5e-9) {
+    printf("%+5.0fns", s * 1e9);
+  } else if (x < 9999.5e-6) {
+    printf("%+5.0fus", s * 1e6);
+  } else if (x < 9999.5e-3) {
+    printf("%+5.0fms", s * 1e3);
+  } else if (x < 999.5) {
+    printf("%+6.1fs", s);
+  } else if (x < 99999.5) {
+    printf("%+6.0fs", s);
+  } else if (x < 99999.5 * 60) {
+    printf("%+6.0fm", s / 60);
+  } else if (x < 99999.5 * 3600) {
+    printf("%+6.0fh", s / 3600);
+  } else if (x < 99999.5 * 3600 * 24) {
+    printf("%+6.0fd", s / (3600 * 24));
   } else {
-    ns = -s * 1e9 + 0.5;
-    ms = -s * 1e3 + 0.5;
-    sign = -1;
+    printf("%+6.0fy", s / (3600 * 24 * 365));
   }
+}
 
-  if (ns <= 9999) {
-    printf("%+5ldns", ns * sign);
-  } else if (ns <= 9999499) {
-    printf("%+5ldus", (ns + 500) / 1000 * sign);
-  } else if (ms <= 9999) {
-    printf("%+5ldms", ms * sign);
+/* ================================================== */
+
+static void
+print_freq_ppm(double f)
+{
+  if (fabs(f) < 99999.5) {
+    printf("%10.3f", f);
   } else {
-    printf("%+6lds", (ms + 500) / 1000 * sign);
+    printf("%10.0f", f);
   }
 }
 
@@ -1614,6 +1736,8 @@ process_cmd_sources(char *line)
               printf("="); break;
             case RPY_SD_MD_REF:
               printf("#"); break;
+            default:
+              printf(" ");
           }
           switch (state) {
             case RPY_SD_ST_SYNC:
@@ -1624,8 +1748,12 @@ process_cmd_sources(char *line)
               printf("x"); break;
             case RPY_SD_ST_JITTERY:
               printf("~"); break;
-            case RPY_SD_ST_OTHER:
+            case RPY_SD_ST_CANDIDATE:
               printf("+"); break;
+            case RPY_SD_ST_OUTLYER:
+              printf("-"); break;
+            default:
+              printf(" ");
           }
 
           printf(" %-25s    %2d   %2d   ", hostname_buf, stratum, poll);
@@ -1660,8 +1788,8 @@ process_cmd_sourcestats(char *line)
 
   char hostname_buf[50];
   unsigned long n_samples, n_runs, span_seconds;
-  double resid_freq_ppm, skew_ppm, sd, est_offset, est_offset_err;
-  unsigned long ref_id;
+  double resid_freq_ppm, skew_ppm, sd, est_offset;
+  uint32_t ref_id;
   IPAddr ip_addr;
 
   verbose = check_for_verbose_flag(line);
@@ -1700,7 +1828,7 @@ process_cmd_sourcestats(char *line)
           skew_ppm = UTI_FloatNetworkToHost(reply.data.sourcestats.skew_ppm);
           sd = UTI_FloatNetworkToHost(reply.data.sourcestats.sd);
           est_offset = UTI_FloatNetworkToHost(reply.data.sourcestats.est_offset);
-          est_offset_err = UTI_FloatNetworkToHost(reply.data.sourcestats.est_offset_err);
+          /* est_offset_err = UTI_FloatNetworkToHost(reply.data.sourcestats.est_offset_err); */
 
           if (ip_addr.family == IPADDR_UNSPEC)
             snprintf(hostname_buf, sizeof(hostname_buf), "%s", UTI_RefidToString(ref_id));
@@ -1711,9 +1839,13 @@ process_cmd_sourcestats(char *line)
             hostname_buf[25] = 0;
           }
 
-          printf("%-25s  %2lu  %2lu  ", hostname_buf, n_samples, n_runs);
+          printf("%-25s %3lu %3lu  ", hostname_buf, n_samples, n_runs);
           print_seconds(span_seconds);
-          printf(" %10.3f %10.3f  ", resid_freq_ppm, skew_ppm);
+          printf(" ");
+          print_freq_ppm(resid_freq_ppm);
+          printf(" ");
+          print_freq_ppm(skew_ppm);
+          printf("  ");
           print_signed_nanoseconds(est_offset);
           printf("  ");
           print_nanoseconds(sd);
@@ -1736,7 +1868,7 @@ process_cmd_tracking(char *line)
   CMD_Request request;
   CMD_Reply reply;
   IPAddr ip_addr;
-  unsigned long ref_id;
+  uint32_t ref_id;
   char host[50];
   char *ref_ip;
   struct timeval ref_time;
@@ -2246,6 +2378,81 @@ process_cmd_activity(const char *line)
 
 /* ================================================== */
 
+static int
+process_cmd_reselectdist(CMD_Request *msg, char *line)
+{
+  double dist;
+  int ok;
+  msg->command = htons(REQ_RESELECTDISTANCE);
+  if (sscanf(line, "%lf", &dist) == 1) {
+    msg->data.reselect_distance.distance = UTI_FloatHostToNetwork(dist);
+    ok = 1;
+  } else {
+    ok = 0;
+  }
+  return ok;
+}
+
+/* ================================================== */
+
+static void
+process_cmd_reselect(CMD_Request *msg, char *line)
+{
+  msg->command = htons(REQ_RESELECT);
+}
+
+/* ================================================== */
+
+static int
+process_cmd_waitsync(char *line)
+{
+  CMD_Request request;
+  CMD_Reply reply;
+  uint32_t ref_id, a, b, c, d;
+  double correction, skew_ppm, max_correction, max_skew_ppm;
+  int ret = 0, max_tries, i;
+
+  max_tries = 0;
+  max_correction = 0.0;
+  max_skew_ppm = 0.0;
+
+  sscanf(line, "%d %lf %lf", &max_tries, &max_correction, &max_skew_ppm);
+
+  request.command = htons(REQ_TRACKING);
+
+  for (i = 1; ; i++) {
+    if (request_reply(&request, &reply, RPY_TRACKING, 0)) {
+      ref_id = ntohl(reply.data.tracking.ref_id);
+      a = (ref_id >> 24);
+      b = (ref_id >> 16) & 0xff;
+      c = (ref_id >> 8) & 0xff;
+      d = (ref_id) & 0xff;
+
+      correction = UTI_FloatNetworkToHost(reply.data.tracking.current_correction);
+      correction = fabs(correction);
+      skew_ppm = UTI_FloatNetworkToHost(reply.data.tracking.skew_ppm);
+
+      printf("try: %d, refid: %d.%d.%d.%d, correction: %.9f, skew: %.3f\n",
+          i, a, b, c, d, correction, skew_ppm);
+
+      if (ref_id != 0 && ref_id != 0x7f7f0101L /* LOCAL refid */ &&
+          (max_correction == 0.0 || correction <= max_correction) &&
+          (max_skew_ppm == 0.0 || skew_ppm <= max_skew_ppm)) {
+        ret = 1;
+      }
+    }
+
+    if (!ret && (!max_tries || i < max_tries)) {
+      sleep(10);
+    } else {
+      break;
+    }
+  }
+  return ret;
+}
+
+/* ================================================== */
+
 static int
 process_cmd_dns(const char *line)
 {
@@ -2268,6 +2475,38 @@ process_cmd_dns(const char *line)
 
 /* ================================================== */
 
+static int
+process_cmd_timeout(const char *line)
+{
+  int timeout;
+
+  timeout = atoi(line);
+  if (timeout < 100) {
+    fprintf(stderr, "Timeout %d is too short\n", timeout);
+    return 0;
+  }
+  initial_timeout = timeout;
+  return 1;
+}
+
+/* ================================================== */
+
+static int
+process_cmd_retries(const char *line)
+{
+  int retries;
+
+  retries = atoi(line);
+  if (retries < 0) {
+    fprintf(stderr, "Invalid maximum number of retries\n");
+    return 0;
+  }
+  max_retries = retries;
+  return 1;
+}
+
+/* ================================================== */
+
 static int
 process_line(char *line, int *quit)
 {
@@ -2305,12 +2544,18 @@ process_line(char *line, int *quit)
     do_normal_submit = process_cmd_maxpoll(&tx_message, p+7);
   } else if (!strncmp(p, "dump", 4)) {
     process_cmd_dump(&tx_message, p+4);
+  } else if (!strncmp(p, "maxdelaydevratio", 16)) {
+    do_normal_submit = process_cmd_maxdelaydevratio(&tx_message, p+16);
   } else if (!strncmp(p, "maxdelayratio", 13)) {
     do_normal_submit = process_cmd_maxdelayratio(&tx_message, p+13);
   } else if (!strncmp(p, "maxdelay", 8)) {
     do_normal_submit = process_cmd_maxdelay(&tx_message, p+8);
   } else if (!strncmp(p, "maxupdateskew", 13)) {
     do_normal_submit = process_cmd_maxupdateskew(&tx_message, p+13);
+  } else if (!strncmp(p, "minstratum", 10)) {
+    do_normal_submit = process_cmd_minstratum(&tx_message, p+10);
+  } else if (!strncmp(p, "polltarget", 10)) {
+    do_normal_submit = process_cmd_polltarget(&tx_message, p+10);
   } else if (!strncmp(p, "settime", 7)) {
     do_normal_submit = 0;
     ret = process_cmd_settime(p+7);
@@ -2381,9 +2626,22 @@ process_line(char *line, int *quit)
   } else if (!strncmp(p, "activity", 8)) {
     ret = process_cmd_activity(p+8);
     do_normal_submit = 0;
+  } else if (!strncmp(p, "reselectdist", 12)) {
+    do_normal_submit = process_cmd_reselectdist(&tx_message, p+12);
+  } else if (!strncmp(p, "reselect", 8)) {
+    process_cmd_reselect(&tx_message, p+8);
+  } else if (!strncmp(p, "waitsync", 8)) {
+    ret = process_cmd_waitsync(p+8);
+    do_normal_submit = 0;
   } else if (!strncmp(p, "dns ", 4)) {
     ret = process_cmd_dns(p+4);
     do_normal_submit = 0;
+  } else if (!strncmp(p, "timeout", 7)) {
+    ret = process_cmd_timeout(p+7);
+    do_normal_submit = 0;
+  } else if (!strncmp(p, "retries", 7)) {
+    ret = process_cmd_retries(p+7);
+    do_normal_submit = 0;
   } else if (!strncmp(p, "help", 4)) {
     do_normal_submit = 0;
     give_help();
@@ -2412,7 +2670,7 @@ process_line(char *line, int *quit)
 /* ================================================== */
 
 static int
-process_args(int argc, char **argv)
+process_args(int argc, char **argv, int multi)
 {
   int total_length, i, ret, quit;
   char *line;
@@ -2423,15 +2681,23 @@ process_args(int argc, char **argv)
   }
 
   line = (char *) malloc((2 + total_length) * sizeof(char));
-  line[0] = 0;
-  for (i=0; i<argc; i++) {
-    strcat(line, argv[i]);
-    if (i + 1 < argc)
-      strcat(line, " ");
-  }
-  strcat(line, "\n");
 
-  ret = process_line(line, &quit);
+  for (i = 0; i < argc; i++) {
+    line[0] = '\0';
+    if (multi) {
+      strcat(line, argv[i]);
+    } else {
+      for (; i < argc; i++) {
+        strcat(line, argv[i]);
+        if (i + 1 < argc)
+          strcat(line, " ");
+      }
+    }
+
+    ret = process_line(line, &quit);
+    if (!ret)
+      break;
+  }
 
   free(line);
 
@@ -2443,12 +2709,12 @@ process_args(int argc, char **argv)
 static void
 display_gpl(void)
 {
-    printf("chrony version %s, copyright (C) 1997-2002 Richard P. Curnow\n"
-           "chrony comes with ABSOLUTELY NO WARRANTY.\n"
-           "This is free software, and you are welcome to redistribute it\n"
-           "under certain conditions.\n"
-           "See the GNU General Public License version 2 for details.\n\n",
-           PROGRAM_VERSION_STRING);
+    printf("chrony version %s\n"
+           "Copyright (C) 1997-2003, 2007, 2009-2011 Richard P. Curnow and others\n"
+           "chrony comes with ABSOLUTELY NO WARRANTY.  This is free software, and\n"
+           "you are welcome to redistribute it under certain conditions.  See the\n"
+           "GNU General Public License version 2 for details.\n\n",
+           CHRONY_VERSION);
 }
 
 /* ================================================== */
@@ -2459,7 +2725,7 @@ main(int argc, char **argv)
   char *line;
   const char *progname = argv[0];
   const char *hostname = "localhost";
-  int quit = 0, ret = 1;
+  int quit = 0, ret = 1, multi = 0;
   int port = DEFAULT_CANDM_PORT;
 
   /* Parse command line options */
@@ -2474,6 +2740,8 @@ main(int argc, char **argv)
       if (*argv) {
         port = atoi(*argv);
       }
+    } else if (!strcmp(*argv, "-m")) {
+      multi = 1;
     } else if (!strcmp(*argv, "-n")) {
       no_dns = 1;
     } else if (!strcmp(*argv, "-4")) {
@@ -2483,10 +2751,10 @@ main(int argc, char **argv)
       DNS_SetAddressFamily(IPADDR_INET6);
       hostname = "::1";
     } else if (!strcmp("-v", *argv) || !strcmp("--version",*argv)) {
-      printf("chronyc (chrony) version %s\n", PROGRAM_VERSION_STRING);
+      printf("chronyc (chrony) version %s\n", CHRONY_VERSION);
       exit(0);
     } else if (!strncmp(*argv, "-", 1)) {
-      fprintf(stderr, "Usage : %s [-h <hostname>] [-p <port-number>] [-n] [-4|-6] [command]\n", progname);
+      fprintf(stderr, "Usage : %s [-h <hostname>] [-p <port-number>] [-n] [-4|-6] [-m] [command]\n", progname);
       exit(1);
     } else {
       break; /* And process remainder of line as a command */
@@ -2504,7 +2772,7 @@ main(int argc, char **argv)
   open_io(hostname, port);
 
   if (argc > 0) {
-    ret = process_args(argc, argv);
+    ret = process_args(argc, argv, multi);
   } else {
     do {
       line = read_line();
