@@ -1,12 +1,9 @@
 /*
-  $Header: /cvs/src/chrony/ntp_sources.c,v 1.18 2003/09/22 21:22:30 richard Exp $
-
-  =======================================================================
-
   chronyd/chronyc - Programs for keeping computer clocks accurate.
 
  **********************************************************************
  * Copyright (C) Richard P. Curnow  1997-2003
+ * Copyright (C) Miroslav Lichvar  2011
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of version 2 of the GNU General Public License as
@@ -30,6 +27,8 @@
 
   */
 
+#include "config.h"
+
 #include "sysincl.h"
 
 #include "ntp_sources.h"
@@ -37,6 +36,9 @@
 #include "util.h"
 #include "logging.h"
 #include "local.h"
+#include "memory.h"
+#include "nameserv.h"
+#include "sched.h"
 
 /* ================================================== */
 
@@ -60,13 +62,25 @@ static int n_sources;
 /* The largest number of sources we want to have stored in the hash table */
 #define MAX_SOURCES 64
 
+/* Source with unknown address (which may be resolved later) */
+struct UnresolvedSource {
+  char *name;
+  int port;
+  NTP_Source_Type type;
+  SourceParameters params;
+  struct UnresolvedSource *next;
+};
+
+static struct UnresolvedSource *unresolved_sources = NULL;
+static int resolving_interval = 0;
+static SCH_TimeoutID resolving_id;
+
 /* ================================================== */
 /* Forward prototypes */
 static void
 slew_sources(struct timeval *raw,
              struct timeval *cooked,
              double dfreq,
-             double afreq,
              double doffset,
              int is_step_change,
              void *anything);
@@ -171,10 +185,9 @@ find_slot(NTP_Remote_Address *remote_addr, int *slot, int *found)
 
 /* ================================================== */
 
-/* Procedure to add a new server source (to which this machine will be
-   a client) */
+/* Procedure to add a new source */
 NSR_Status
-NSR_AddServer(NTP_Remote_Address *remote_addr, SourceParameters *params)
+NSR_AddSource(NTP_Remote_Address *remote_addr, NTP_Source_Type type, SourceParameters *params)
 {
   int slot, found;
 
@@ -196,7 +209,7 @@ NSR_AddServer(NTP_Remote_Address *remote_addr, SourceParameters *params)
       return NSR_InvalidAF;
     } else {
       n_sources++;
-      records[slot].data = NCR_GetServerInstance(remote_addr, params); /* Will need params passing through */
+      records[slot].data = NCR_GetInstance(remote_addr, type, params); /* Will need params passing through */
       records[slot].remote_addr = NCR_GetRemoteAddress(records[slot].data);
       return NSR_Success;
     }
@@ -205,34 +218,70 @@ NSR_AddServer(NTP_Remote_Address *remote_addr, SourceParameters *params)
 
 /* ================================================== */
 
-/* Procedure to add a new peer. */
-NSR_Status
-NSR_AddPeer(NTP_Remote_Address *remote_addr, SourceParameters *params)
+static void
+resolve_sources(void *arg)
 {
-  int slot, found;
+  NTP_Remote_Address address;
+  struct UnresolvedSource *us, **i;
+  DNS_Status s;
 
-  assert(initialised);
+  memset(&address.local_ip_addr, 0, sizeof (address.local_ip_addr));
 
-#if 0
-  LOG(LOGS_INFO, LOGF_NtpSources, "IP=%s port=%d", UTI_IPToString(&remote_addr->ip_addr), remote_addr->port);
-#endif
+  DNS_Reload();
 
-  /* Find empty bin & check that we don't have the address already */
-  find_slot(remote_addr, &slot, &found);
-  if (found) {
-    return NSR_AlreadyInUse;
-  } else {
-    if (n_sources == MAX_SOURCES) {
-      return NSR_TooManySources;
-    } else if (remote_addr->ip_addr.family != IPADDR_INET4 &&
-               remote_addr->ip_addr.family != IPADDR_INET6) {
-      return NSR_InvalidAF;
+  for (i = &unresolved_sources; *i; ) {
+    us = *i;
+    s = DNS_Name2IPAddress(us->name, &address.ip_addr);
+    if (s == DNS_TryAgain) {
+      i = &(*i)->next;
+      continue;
+    } else if (s == DNS_Success) {
+      address.port = us->port;
+      NSR_AddSource(&address, us->type, &us->params);
     } else {
-      n_sources++;
-      records[slot].data = NCR_GetPeerInstance(remote_addr, params); /* Will need params passing through */
-      records[slot].remote_addr = NCR_GetRemoteAddress(records[slot].data);
-      return NSR_Success;
+      LOG(LOGS_WARN, LOGF_NtpSources, "Invalid host %s", us->name);
     }
+    
+    *i = us->next;
+
+    Free(us->name);
+    Free(us);
+  }
+
+  if (unresolved_sources) {
+    /* Try again later */
+    if (resolving_interval < 9)
+      resolving_interval++;
+    resolving_id = SCH_AddTimeoutByDelay(7 * (1 << resolving_interval), resolve_sources, NULL);
+  } else {
+    resolving_interval = 0;
+  }
+}
+
+/* ================================================== */
+
+/* Procedure to add a new server or peer source, but instead of an IP address
+   only a name is provided */
+void
+NSR_AddUnresolvedSource(char *name, int port, NTP_Source_Type type, SourceParameters *params)
+{
+  struct UnresolvedSource *us, **i;
+
+  us = MallocNew(struct UnresolvedSource);
+
+  us->name = name;
+  us->port = port;
+  us->type = type;
+  us->params = *params;
+  us->next = NULL;
+
+  for (i = &unresolved_sources; *i; i = &(*i)->next)
+    ;
+  *i = us;
+
+  if (!resolving_interval) {
+    resolving_interval = 2;
+    resolving_id = SCH_AddTimeoutByDelay(7 * (1 << resolving_interval), resolve_sources, NULL);
   }
 }
 
@@ -245,26 +294,48 @@ NSR_AddPeer(NTP_Remote_Address *remote_addr, SourceParameters *params)
 NSR_Status
 NSR_RemoveSource(NTP_Remote_Address *remote_addr)
 {
-  int slot, found;
+  int i, slot, found;
+  SourceRecord temp_records[N_RECORDS];
 
   assert(initialised);
 
   find_slot(remote_addr, &slot, &found);
   if (!found) {
     return NSR_NoSuchSource;
-  } else {
-    n_sources--;
-    records[slot].remote_addr = NULL;
-    NCR_DestroyInstance(records[slot].data);
-    return NSR_Success;
   }
+
+  n_sources--;
+  records[slot].remote_addr = NULL;
+  NCR_DestroyInstance(records[slot].data);
+
+  /* Rehash the table to make sure there are no broken probe sequences.
+     This is costly, but it's not expected to happen frequently. */
+
+  memcpy(temp_records, records, sizeof (records));
+
+  for (i = 0; i < N_RECORDS; i++) {
+    records[i].remote_addr = NULL;
+  }
+  
+  for (i = 0; i < N_RECORDS; i++) {
+    if (!temp_records[i].remote_addr)
+      continue;
+
+    find_slot(temp_records[i].remote_addr, &slot, &found);
+    assert(!found);
+
+    records[slot].remote_addr = temp_records[i].remote_addr;
+    records[slot].data = temp_records[i].data;
+  }
+
+  return NSR_Success;
 }
 
 /* ================================================== */
 
 /* This routine is called by ntp_io when a new packet arrives off the network.*/
 void
-NSR_ProcessReceive(NTP_Packet *message, struct timeval *now, NTP_Remote_Address *remote_addr)
+NSR_ProcessReceive(NTP_Packet *message, struct timeval *now, double now_err, NTP_Remote_Address *remote_addr)
 {
   int slot, found;
 
@@ -278,9 +349,9 @@ NSR_ProcessReceive(NTP_Packet *message, struct timeval *now, NTP_Remote_Address
   
   find_slot(remote_addr, &slot, &found);
   if (found == 2) { /* Must match IP address AND port number */
-    NCR_ProcessNoauthKnown(message, now, records[slot].data);
+    NCR_ProcessNoauthKnown(message, now, now_err, records[slot].data);
   } else {
-    NCR_ProcessNoauthUnknown(message, now, remote_addr);
+    NCR_ProcessNoauthUnknown(message, now, now_err, remote_addr);
   }
 }
 
@@ -288,7 +359,7 @@ NSR_ProcessReceive(NTP_Packet *message, struct timeval *now, NTP_Remote_Address
 
 /* This routine is called by ntp_io when a new packet with an authentication tail arrives off the network */
 void
-NSR_ProcessAuthenticatedReceive(NTP_Packet *message, struct timeval *now, NTP_Remote_Address *remote_addr)
+NSR_ProcessAuthenticatedReceive(NTP_Packet *message, struct timeval *now, double now_err, NTP_Remote_Address *remote_addr)
 {
   int slot, found;
 
@@ -296,9 +367,9 @@ NSR_ProcessAuthenticatedReceive(NTP_Packet *message, struct timeval *now, NTP_Re
 
   find_slot(remote_addr, &slot, &found);
   if (found == 2) {
-    NCR_ProcessAuthKnown(message, now, records[slot].data);
+    NCR_ProcessAuthKnown(message, now, now_err, records[slot].data);
   } else {
-    NCR_ProcessAuthUnknown(message, now, remote_addr);
+    NCR_ProcessAuthUnknown(message, now, now_err, remote_addr);
   }
 }
 
@@ -308,7 +379,6 @@ static void
 slew_sources(struct timeval *raw,
              struct timeval *cooked,
              double dfreq,
-             double afreq,
              double doffset,
              int is_step_change,
              void *anything)
@@ -336,6 +406,13 @@ NSR_TakeSourcesOnline(IPAddr *mask, IPAddr *address)
   int i;
   int any;
 
+  /* Try to resolve unresolved sources now */
+  if (resolving_interval) {
+    SCH_RemoveTimeout(resolving_id);
+    resolving_interval--;
+    resolve_sources(NULL);
+  }
+
   any = 0;
   for (i=0; i<N_RECORDS; i++) {
     if (records[i].remote_addr) {
@@ -347,6 +424,15 @@ NSR_TakeSourcesOnline(IPAddr *mask, IPAddr *address)
     }
   }
 
+  if (address->family == IPADDR_UNSPEC) {
+    struct UnresolvedSource *us;
+
+    for (us = unresolved_sources; us; us = us->next) {
+      any = 1;
+      us->params.online = 1;
+    }
+  }
+
   return any;
 }
 
@@ -355,20 +441,38 @@ NSR_TakeSourcesOnline(IPAddr *mask, IPAddr *address)
 int
 NSR_TakeSourcesOffline(IPAddr *mask, IPAddr *address)
 {
-  int i;
-  int any;
+  int i, any, syncpeer;
 
   any = 0;
+  syncpeer = -1;
   for (i=0; i<N_RECORDS; i++) {
     if (records[i].remote_addr) {
       if (address->family == IPADDR_UNSPEC ||
           !UTI_CompareIPs(&records[i].remote_addr->ip_addr, address, mask)) {
         any = 1;
+        if (NCR_IsSyncPeer(records[i].data)) {
+          syncpeer = i;
+          continue;
+        }
         NCR_TakeSourceOffline(records[i].data);
       }
     }
   }
 
+  /* Take sync peer offline as last to avoid reference switching */
+  if (syncpeer >= 0) {
+    NCR_TakeSourceOffline(records[syncpeer].data);
+  }
+
+  if (address->family == IPADDR_UNSPEC) {
+    struct UnresolvedSource *us;
+
+    for (us = unresolved_sources; us; us = us->next) {
+      any = 1;
+      us->params.online = 0;
+    }
+  }
+
   return any;
 }
 
@@ -450,6 +554,63 @@ NSR_ModifyMaxdelayratio(IPAddr *address, double new_max_delay_ratio)
 
 /* ================================================== */
 
+int
+NSR_ModifyMaxdelaydevratio(IPAddr *address, double new_max_delay_dev_ratio)
+{
+  int slot, found;
+  NTP_Remote_Address addr;
+  addr.ip_addr = *address;
+  addr.port = 0;
+
+  find_slot(&addr, &slot, &found);
+  if (found == 0) {
+    return 0;
+  } else {
+    NCR_ModifyMaxdelaydevratio(records[slot].data, new_max_delay_dev_ratio);
+    return 1;
+  }
+}
+
+/* ================================================== */
+
+int
+NSR_ModifyMinstratum(IPAddr *address, int new_min_stratum)
+{
+  int slot, found;
+  NTP_Remote_Address addr;
+  addr.ip_addr = *address;
+  addr.port = 0;
+
+  find_slot(&addr, &slot, &found);
+  if (found == 0) {
+    return 0;
+  } else {
+    NCR_ModifyMinstratum(records[slot].data, new_min_stratum);
+    return 1;
+  }
+}
+
+/* ================================================== */
+
+int
+NSR_ModifyPolltarget(IPAddr *address, int new_poll_target)
+{
+  int slot, found;
+  NTP_Remote_Address addr;
+  addr.ip_addr = *address;
+  addr.port = 0;
+
+  find_slot(&addr, &slot, &found);
+  if (found == 0) {
+    return 0;
+  } else {
+    NCR_ModifyPolltarget(records[slot].data, new_poll_target);
+    return 1;
+  }
+}
+
+/* ================================================== */
+
 int
 NSR_InitiateSampleBurst(int n_good_samples, int n_total_samples,
                         IPAddr *mask, IPAddr *address)
@@ -499,6 +660,7 @@ void
 NSR_GetActivityReport(RPT_ActivityReport *report)
 {
   int i;
+  struct UnresolvedSource *us;
 
   report->online = 0;
   report->offline = 0;
@@ -511,6 +673,12 @@ NSR_GetActivityReport(RPT_ActivityReport *report)
                                     &report->burst_online, &report->burst_offline);
     }
   }
+
+  /* Add unresolved sources to offline count */
+  for (us = unresolved_sources; us; us = us->next) {
+    report->offline++;
+  }
+
   return;
 }
 
