@@ -2,7 +2,7 @@
   chronyd/chronyc - Programs for keeping computer clocks accurate.
 
  **********************************************************************
- * Copyright (C) Miroslav Lichvar  2009
+ * Copyright (C) Miroslav Lichvar  2009-2011
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of version 2 of the GNU General Public License as
@@ -25,6 +25,8 @@
 
   */
 
+#include "config.h"
+
 #include "refclock.h"
 #include "reference.h"
 #include "conf.h"
@@ -33,8 +35,8 @@
 #include "util.h"
 #include "sources.h"
 #include "logging.h"
+#include "regress.h"
 #include "sched.h"
-#include "mkdirpp.h"
 
 /* list of refclock drivers */
 extern RefclockDriver RCL_SHM_driver;
@@ -43,6 +45,7 @@ extern RefclockDriver RCL_PPS_driver;
 
 struct FilterSample {
   double offset;
+  double dispersion;
   struct timeval sample_time;
 };
 
@@ -51,7 +54,13 @@ struct MedianFilter {
   int index;
   int used;
   int last;
+  int avg_var_n;
+  double avg_var;
   struct FilterSample *samples;
+  int *selected;
+  double *x_data;
+  double *y_data;
+  double *w_data;
 };
 
 struct RCL_Instance_Record {
@@ -62,14 +71,14 @@ struct RCL_Instance_Record {
   int driver_poll;
   int driver_polled;
   int poll;
-  int missed_samples;
   int leap_status;
   int pps_rate;
   struct MedianFilter filter;
-  unsigned long ref_id;
-  unsigned long lock_ref;
+  uint32_t ref_id;
+  uint32_t lock_ref;
   double offset;
   double delay;
+  double precision;
   SCH_TimeoutID timeout_id;
   SRC_Instance source;
 };
@@ -79,42 +88,35 @@ struct RCL_Instance_Record {
 static struct RCL_Instance_Record refclocks[MAX_RCL_SOURCES];
 static int n_sources = 0;
 
-#define REFCLOCKS_LOG "refclocks.log"
-static FILE *logfile = NULL;
-static char *logfilename = NULL;
-static unsigned long logwrites = 0;
+static LOG_FileID logfileid;
 
 static int valid_sample_time(RCL_Instance instance, struct timeval *tv);
 static int pps_stratum(RCL_Instance instance, struct timeval *tv);
 static void poll_timeout(void *arg);
-static void slew_samples(struct timeval *raw, struct timeval *cooked, double dfreq, double afreq,
+static void slew_samples(struct timeval *raw, struct timeval *cooked, double dfreq,
              double doffset, int is_step_change, void *anything);
-static void log_sample(RCL_Instance instance, struct timeval *sample_time, int pulse, double raw_offset, double cooked_offset);
+static void add_dispersion(double dispersion, void *anything);
+static void log_sample(RCL_Instance instance, struct timeval *sample_time, int filtered, int pulse, double raw_offset, double cooked_offset, double dispersion);
 
 static void filter_init(struct MedianFilter *filter, int length);
 static void filter_fini(struct MedianFilter *filter);
 static void filter_reset(struct MedianFilter *filter);
-static void filter_add_sample(struct MedianFilter *filter, struct timeval *sample_time, double offset);
-static int filter_get_last_sample(struct MedianFilter *filter, struct timeval *sample_time, double *offset);
+static double filter_get_avg_sample_dispersion(struct MedianFilter *filter);
+static void filter_add_sample(struct MedianFilter *filter, struct timeval *sample_time, double offset, double dispersion);
+static int filter_get_last_sample(struct MedianFilter *filter, struct timeval *sample_time, double *offset, double *dispersion);
+static int filter_select_samples(struct MedianFilter *filter);
 static int filter_get_sample(struct MedianFilter *filter, struct timeval *sample_time, double *offset, double *dispersion);
 static void filter_slew_samples(struct MedianFilter *filter, struct timeval *when, double dfreq, double doffset);
+static void filter_add_dispersion(struct MedianFilter *filter, double dispersion);
 
 void
 RCL_Initialise(void)
 {
   CNF_AddRefclocks();
 
-  if (CNF_GetLogRefclocks()) {
-    char *logdir = CNF_GetLogDir();
-    if (!mkdir_and_parents(logdir)) {
-      LOG(LOGS_ERR, LOGF_Refclock, "Could not create directory %s", logdir);
-    } else {
-      logfilename = MallocArray(char, 2 + strlen(logdir) + strlen(REFCLOCKS_LOG));
-      strcpy(logfilename, logdir);
-      strcat(logfilename, "/");
-      strcat(logfilename, REFCLOCKS_LOG);
-    }
-  }
+  logfileid = CNF_GetLogRefclocks() ? LOG_FileOpen("refclocks",
+      "   Date (UTC) Time         Refid  DP L P  Raw offset   Cooked offset      Disp.")
+    : -1;
 }
 
 void
@@ -132,12 +134,10 @@ RCL_Finalise(void)
     Free(inst->driver_parameter);
   }
 
-  if (n_sources > 0)
+  if (n_sources > 0) {
     LCL_RemoveParameterChangeHandler(slew_samples, NULL);
-
-  if (logfile)
-    fclose(logfile);
-  Free(logfilename);
+    LCL_RemoveDispersionNotifyHandler(add_dispersion, NULL);
+  }
 }
 
 int
@@ -150,13 +150,16 @@ RCL_AddRefclock(RefclockParameters *params)
   if (n_sources == MAX_RCL_SOURCES)
     return 0;
 
-  if (strncmp(params->driver_name, "SHM", 4) == 0) {
+  if (strcmp(params->driver_name, "SHM") == 0) {
     inst->driver = &RCL_SHM_driver;
-  } else if (strncmp(params->driver_name, "SOCK", 4) == 0) {
+    inst->precision = 1e-6;
+  } else if (strcmp(params->driver_name, "SOCK") == 0) {
     inst->driver = &RCL_SOCK_driver;
+    inst->precision = 1e-9;
     pps_source = 1;
-  } else if (strncmp(params->driver_name, "PPS", 4) == 0) {
+  } else if (strcmp(params->driver_name, "PPS") == 0) {
     inst->driver = &RCL_PPS_driver;
+    inst->precision = 1e-9;
     pps_source = 1;
   } else {
     LOG_FATAL(LOGF_Refclock, "unknown refclock driver %s", params->driver_name);
@@ -173,13 +176,14 @@ RCL_AddRefclock(RefclockParameters *params)
   inst->driver_parameter_length = 0;
   inst->driver_poll = params->driver_poll;
   inst->poll = params->poll;
-  inst->missed_samples = 0;
   inst->driver_polled = 0;
-  inst->leap_status = 0;
+  inst->leap_status = LEAP_Normal;
   inst->pps_rate = params->pps_rate;
   inst->lock_ref = params->lock_ref_id;
   inst->offset = params->offset;
   inst->delay = params->delay;
+  if (params->precision > 0.0)
+    inst->precision = params->precision;
   inst->timeout_id = -1;
   inst->source = NULL;
 
@@ -199,18 +203,31 @@ RCL_AddRefclock(RefclockParameters *params)
     inst->pps_rate = 0;
   }
 
-  if (inst->driver_poll > inst->poll)
-    inst->driver_poll = inst->poll;
-
   if (params->ref_id)
     inst->ref_id = params->ref_id;
   else {
     unsigned char ref[5] = { 0, 0, 0, 0, 0 };
 
-    snprintf((char *)ref, 5, "%3s%d", params->driver_name, n_sources % 10);
+    snprintf((char *)ref, 5, "%3.3s%d", params->driver_name, n_sources % 10);
     inst->ref_id = ref[0] << 24 | ref[1] << 16 | ref[2] << 8 | ref[3];
   }
 
+  if (inst->driver->poll) {
+    int max_samples;
+
+    if (inst->driver_poll > inst->poll)
+      inst->driver_poll = inst->poll;
+
+    max_samples = 1 << (inst->poll - inst->driver_poll);
+    if (max_samples < params->filter_length) {
+      if (max_samples < 4) {
+        LOG(LOGS_WARN, LOGF_Refclock, "Setting filter length for %s to %d",
+            UTI_RefidToString(inst->ref_id), max_samples);
+      }
+      params->filter_length = max_samples;
+    }
+  }
+
   if (inst->driver->init)
     if (!inst->driver->init(inst)) {
       LOG_FATAL(LOGF_Refclock, "refclock %s initialisation failed", params->driver_name);
@@ -219,12 +236,16 @@ RCL_AddRefclock(RefclockParameters *params)
 
   filter_init(&inst->filter, params->filter_length);
 
+  inst->source = SRC_CreateNewInstance(inst->ref_id, SRC_REFCLOCK, params->sel_option, NULL);
+
 #if 0
   LOG(LOGS_INFO, LOGF_Refclock, "refclock added poll=%d dpoll=%d filter=%d",
 		  inst->poll, inst->driver_poll, params->filter_length);
 #endif
   n_sources++;
 
+  Free(params->driver_name);
+
   return 1;
 }
 
@@ -236,7 +257,7 @@ RCL_StartRefclocks(void)
   for (i = 0; i < n_sources; i++) {
     RCL_Instance inst = &refclocks[i];
 
-    inst->source = SRC_CreateNewInstance(inst->ref_id, SRC_REFCLOCK, NULL);
+    SRC_SetSelectable(inst->source);
     inst->timeout_id = SCH_AddTimeoutByDelay(0.0, poll_timeout, (void *)inst);
 
     if (inst->lock_ref) {
@@ -248,15 +269,17 @@ RCL_StartRefclocks(void)
       inst->lock_ref = -1;
   }
 
-  if (n_sources > 0)
+  if (n_sources > 0) {
     LCL_AddParameterChangeHandler(slew_samples, NULL);
+    LCL_AddDispersionNotifyHandler(add_dispersion, NULL);
+  }
 }
 
 void
 RCL_ReportSource(RPT_SourceReport *report, struct timeval *now)
 {
   int i;
-  unsigned long ref_id;
+  uint32_t ref_id;
 
   assert(report->ip_addr.family == IPADDR_INET4);
   ref_id = report->ip_addr.addr.in4;
@@ -315,26 +338,36 @@ RCL_GetDriverOption(RCL_Instance instance, char *name)
 }
 
 int
-RCL_AddSample(RCL_Instance instance, struct timeval *sample_time, double offset, NTP_Leap leap_status)
+RCL_AddSample(RCL_Instance instance, struct timeval *sample_time, double offset, int leap)
 {
-  double correction;
+  double correction, dispersion;
   struct timeval cooked_time;
 
-  correction = LCL_GetOffsetCorrection(sample_time);
+  LCL_GetOffsetCorrection(sample_time, &correction, &dispersion);
   UTI_AddDoubleToTimeval(sample_time, correction, &cooked_time);
+  dispersion += instance->precision + filter_get_avg_sample_dispersion(&instance->filter);
 
   if (!valid_sample_time(instance, sample_time))
     return 0;
 
-#if 0
-  LOG(LOGS_INFO, LOGF_Refclock, "refclock sample offset=%.9f cooked=%.9f",
-      offset, offset - correction + instance->offset);
-#endif
+  filter_add_sample(&instance->filter, &cooked_time, offset - correction + instance->offset, dispersion);
 
-  filter_add_sample(&instance->filter, &cooked_time, offset - correction + instance->offset);
-  instance->leap_status = leap_status;
+  switch (leap) {
+    case LEAP_Normal:
+    case LEAP_InsertSecond:
+    case LEAP_DeleteSecond:
+      instance->leap_status = leap;
+      break;
+    default:
+      instance->leap_status = LEAP_Unsynchronised;
+      break;
+  }
 
-  log_sample(instance, &cooked_time, 0, offset, offset - correction + instance->offset);
+  log_sample(instance, &cooked_time, 0, 0, offset, offset - correction + instance->offset, dispersion);
+
+  /* for logging purposes */
+  if (!instance->driver->poll)
+    instance->driver_polled++;
 
   return 1;
 }
@@ -342,18 +375,13 @@ RCL_AddSample(RCL_Instance instance, struct timeval *sample_time, double offset,
 int
 RCL_AddPulse(RCL_Instance instance, struct timeval *pulse_time, double second)
 {
-  double correction, offset;
+  double correction, dispersion, offset;
   struct timeval cooked_time;
   int rate;
 
-  struct timeval ref_time;
-  int is_synchronised, stratum;
-  double root_delay, root_dispersion, distance;
-  NTP_Leap leap;
-  unsigned long ref_id;
-
-  correction = LCL_GetOffsetCorrection(pulse_time);
+  LCL_GetOffsetCorrection(pulse_time, &correction, &dispersion);
   UTI_AddDoubleToTimeval(pulse_time, correction, &cooked_time);
+  dispersion += instance->precision + filter_get_avg_sample_dispersion(&instance->filter);
 
   if (!valid_sample_time(instance, pulse_time))
     return 0;
@@ -372,10 +400,10 @@ RCL_AddPulse(RCL_Instance instance, struct timeval *pulse_time, double second)
 
   if (instance->lock_ref != -1) {
     struct timeval ref_sample_time;
-    double sample_diff, ref_offset, shift;
+    double sample_diff, ref_offset, ref_dispersion, shift;
 
     if (!filter_get_last_sample(&refclocks[instance->lock_ref].filter,
-          &ref_sample_time, &ref_offset))
+          &ref_sample_time, &ref_offset, &ref_dispersion))
       return 0;
 
     UTI_DiffTimevalsToDouble(&sample_diff, &cooked_time, &ref_sample_time);
@@ -390,7 +418,7 @@ RCL_AddPulse(RCL_Instance instance, struct timeval *pulse_time, double second)
 
     offset += shift;
 
-    if (fabs(ref_offset - offset) >= 0.2 / rate)
+    if (fabs(ref_offset - offset) + ref_dispersion + dispersion >= 0.2 / rate)
       return 0;
 
 #if 0
@@ -398,6 +426,12 @@ RCL_AddPulse(RCL_Instance instance, struct timeval *pulse_time, double second)
         second, offset, ref_offset - offset, sample_diff);
 #endif
   } else {
+    struct timeval ref_time;
+    int is_synchronised, stratum;
+    double root_delay, root_dispersion, distance;
+    NTP_Leap leap;
+    uint32_t ref_id;
+
     /* Ignore the pulse if we are not well synchronized */
 
     REF_GetReferenceParams(&cooked_time, &is_synchronised, &leap, &stratum,
@@ -415,29 +449,18 @@ RCL_AddPulse(RCL_Instance instance, struct timeval *pulse_time, double second)
     }
   }
 
-#if 0
-  LOG(LOGS_INFO, LOGF_Refclock, "refclock pulse second=%.9f offset=%.9f",
-      second, offset);
-#endif
-
-  filter_add_sample(&instance->filter, &cooked_time, offset);
+  filter_add_sample(&instance->filter, &cooked_time, offset, dispersion);
   instance->leap_status = LEAP_Normal;
 
-  log_sample(instance, &cooked_time, 1, second, offset);
+  log_sample(instance, &cooked_time, 0, 1, second, offset, dispersion);
+
+  /* for logging purposes */
+  if (!instance->driver->poll)
+    instance->driver_polled++;
 
   return 1;
 }
 
-void
-RCL_CycleLogFile(void)
-{
-  if (logfile) {
-    fclose(logfile);
-    logfile = NULL;
-    logwrites = 0;
-  }
-}
-
 static int
 valid_sample_time(RCL_Instance instance, struct timeval *tv)
 {
@@ -458,7 +481,7 @@ pps_stratum(RCL_Instance instance, struct timeval *tv)
   int is_synchronised, stratum, i;
   double root_delay, root_dispersion;
   NTP_Leap leap;
-  unsigned long ref_id;
+  uint32_t ref_id;
 
   REF_GetReferenceParams(tv, &is_synchronised, &leap, &stratum,
       &ref_id, &ref_time, &root_delay, &root_dispersion);
@@ -470,7 +493,8 @@ pps_stratum(RCL_Instance instance, struct timeval *tv)
 
   /* Or the current source is another PPS refclock */ 
   for (i = 0; i < n_sources; i++) {
-    if (refclocks[i].ref_id == ref_id && refclocks[i].pps_rate)
+    if (refclocks[i].ref_id == ref_id &&
+        refclocks[i].pps_rate && refclocks[i].lock_ref == -1)
       return stratum - 1;
   }
 
@@ -499,29 +523,22 @@ poll_timeout(void *arg)
     int sample_ok, stratum;
 
     sample_ok = filter_get_sample(&inst->filter, &sample_time, &offset, &dispersion);
-    filter_reset(&inst->filter);
     inst->driver_polled = 0;
 
     if (sample_ok) {
-#if 0
-      LOG(LOGS_INFO, LOGF_Refclock, "refclock filtered sample: offset=%.9f dispersion=%.9f [%s]",
-          offset, dispersion, UTI_TimevalToString(&sample_time));
-#endif
-
-      if (inst->pps_rate)
+      if (inst->pps_rate && inst->lock_ref == -1)
         /* Handle special case when PPS is used with local stratum */
         stratum = pps_stratum(inst, &sample_time);
       else
         stratum = 0;
 
-      SRC_SetReachable(inst->source);
+      SRC_UpdateReachability(inst->source, 1);
       SRC_AccumulateSample(inst->source, &sample_time, offset,
           inst->delay, dispersion, inst->delay, dispersion, stratum, inst->leap_status);
-      inst->missed_samples = 0;
+
+      log_sample(inst, &sample_time, 1, 0, 0.0, offset, dispersion);
     } else {
-      inst->missed_samples++;
-      if (inst->missed_samples > 9)
-        SRC_UnsetReachable(inst->source);
+      SRC_UpdateReachability(inst->source, 0);
     }
   }
 
@@ -534,7 +551,7 @@ poll_timeout(void *arg)
 }
 
 static void
-slew_samples(struct timeval *raw, struct timeval *cooked, double dfreq, double afreq,
+slew_samples(struct timeval *raw, struct timeval *cooked, double dfreq,
              double doffset, int is_step_change, void *anything)
 {
   int i;
@@ -544,30 +561,24 @@ slew_samples(struct timeval *raw, struct timeval *cooked, double dfreq, double a
 }
 
 static void
-log_sample(RCL_Instance instance, struct timeval *sample_time, int pulse, double raw_offset, double cooked_offset)
+add_dispersion(double dispersion, void *anything)
+{
+  int i;
+
+  for (i = 0; i < n_sources; i++)
+    filter_add_dispersion(&refclocks[i].filter, dispersion);
+}
+
+static void
+log_sample(RCL_Instance instance, struct timeval *sample_time, int filtered, int pulse, double raw_offset, double cooked_offset, double dispersion)
 {
   char sync_stats[4] = {'N', '+', '-', '?'};
 
-  if (!logfilename)
+  if (logfileid == -1)
     return;
 
-  if (!logfile) {
-      logfile = fopen(logfilename, "a");
-      if (!logfile) {
-        LOG(LOGS_WARN, LOGF_Refclock, "Couldn't open logfile %s for update", logfilename);
-        Free(logfilename);
-        logfilename = NULL;
-        return;
-      }
-  }
-
-  if (((logwrites++) % 32) == 0) {
-    fprintf(logfile,
-        "====================================================================\n"
-        "   Date (UTC) Time         Refid  DP L P  Raw offset   Cooked offset\n"
-        "====================================================================\n");
-  }
-  fprintf(logfile, "%s.%06d %-5s %3d %1c %1d %13.6e %13.6e\n",
+  if (!filtered) {
+    LOG_FileWrite(logfileid, "%s.%06d %-5s %3d %1c %1d %13.6e %13.6e %10.3e",
       UTI_TimeToLogForm(sample_time->tv_sec),
       (int)sample_time->tv_usec,
       UTI_RefidToString(instance->ref_id),
@@ -575,8 +586,17 @@ log_sample(RCL_Instance instance, struct timeval *sample_time, int pulse, double
       sync_stats[instance->leap_status],
       pulse,
       raw_offset,
-      cooked_offset);
-  fflush(logfile);
+      cooked_offset,
+      dispersion);
+  } else {
+    LOG_FileWrite(logfileid, "%s.%06d %-5s   - %1c -       -       %13.6e %10.3e",
+      UTI_TimeToLogForm(sample_time->tv_sec),
+      (int)sample_time->tv_usec,
+      UTI_RefidToString(instance->ref_id),
+      sync_stats[instance->leap_status],
+      cooked_offset,
+      dispersion);
+  }
 }
 
 static void
@@ -589,13 +609,24 @@ filter_init(struct MedianFilter *filter, int length)
   filter->index = -1;
   filter->used = 0;
   filter->last = -1;
+  /* set first estimate to system precision */
+  filter->avg_var_n = 0;
+  filter->avg_var = LCL_GetSysPrecisionAsQuantum() * LCL_GetSysPrecisionAsQuantum();
   filter->samples = MallocArray(struct FilterSample, filter->length);
+  filter->selected = MallocArray(int, filter->length);
+  filter->x_data = MallocArray(double, filter->length);
+  filter->y_data = MallocArray(double, filter->length);
+  filter->w_data = MallocArray(double, filter->length);
 }
 
 static void
 filter_fini(struct MedianFilter *filter)
 {
   Free(filter->samples);
+  Free(filter->selected);
+  Free(filter->x_data);
+  Free(filter->y_data);
+  Free(filter->w_data);
 }
 
 static void
@@ -605,8 +636,14 @@ filter_reset(struct MedianFilter *filter)
   filter->used = 0;
 }
 
+static double
+filter_get_avg_sample_dispersion(struct MedianFilter *filter)
+{
+  return sqrt(filter->avg_var);
+}
+
 static void
-filter_add_sample(struct MedianFilter *filter, struct timeval *sample_time, double offset)
+filter_add_sample(struct MedianFilter *filter, struct timeval *sample_time, double offset, double dispersion)
 {
   filter->index++;
   filter->index %= filter->length;
@@ -616,23 +653,30 @@ filter_add_sample(struct MedianFilter *filter, struct timeval *sample_time, doub
 
   filter->samples[filter->index].sample_time = *sample_time;
   filter->samples[filter->index].offset = offset;
+  filter->samples[filter->index].dispersion = dispersion;
 }
 
 static int
-filter_get_last_sample(struct MedianFilter *filter, struct timeval *sample_time, double *offset)
+filter_get_last_sample(struct MedianFilter *filter, struct timeval *sample_time, double *offset, double *dispersion)
 {
   if (filter->last < 0)
     return 0;
 
   *sample_time = filter->samples[filter->last].sample_time;
   *offset = filter->samples[filter->last].offset;
+  *dispersion = filter->samples[filter->last].dispersion;
   return 1;
 }
 
+static const struct FilterSample *tmp_sorted_array;
+
 static int
 sample_compare(const void *a, const void *b)
 {
-  const struct FilterSample *s1 = a, *s2 = b;
+  const struct FilterSample *s1, *s2;
+
+  s1 = &tmp_sorted_array[*(int *)a];
+  s2 = &tmp_sorted_array[*(int *)b];
 
   if (s1->offset < s2->offset)
     return -1;
@@ -641,55 +685,184 @@ sample_compare(const void *a, const void *b)
   return 0;
 }
 
+int
+filter_select_samples(struct MedianFilter *filter)
+{
+  int i, j, k, o, from, to, *selected;
+  double min_dispersion;
+
+  if (filter->used < 1)
+    return 0;
+
+  /* for lengths below 4 require full filter,
+     for 4 and above require at least 4 samples */
+  if ((filter->length < 4 && filter->used != filter->length) ||
+      (filter->length >= 4 && filter->used < 4))
+    return 0;
+
+  selected = filter->selected;
+
+  if (filter->used > 4) {
+    /* select samples with dispersion better than 1.5 * minimum */
+
+    for (i = 1, min_dispersion = filter->samples[0].dispersion; i < filter->used; i++) {
+      if (min_dispersion > filter->samples[i].dispersion)
+        min_dispersion = filter->samples[i].dispersion;
+    }
+
+    for (i = j = 0; i < filter->used; i++) {
+      if (filter->samples[i].dispersion <= 1.5 * min_dispersion)
+        selected[j++] = i;
+    }
+  } else {
+    j = 0;
+  }
+
+  if (j < 4) {
+    /* select all samples */
+
+    for (j = 0; j < filter->used; j++)
+      selected[j] = j;
+  }
+
+  /* and sort their indices by offset */
+  tmp_sorted_array = filter->samples;
+  qsort(selected, j, sizeof (int), sample_compare);
+
+  /* select 60 percent of the samples closest to the median */ 
+  if (j > 2) {
+    from = j / 5;
+    if (from < 1)
+      from = 1;
+    to = j - from;
+  } else {
+    from = 0;
+    to = j;
+  }
+
+  /* mark unused samples and sort the rest from oldest to newest */
+
+  o = filter->used - filter->index - 1;
+
+  for (i = 0; i < from; i++)
+    selected[i] = -1;
+  for (; i < to; i++)
+    selected[i] = (selected[i] + o) % filter->used;
+  for (; i < filter->used; i++)
+    selected[i] = -1;
+
+  for (i = from; i < to; i++) {
+    j = selected[i];
+    selected[i] = -1;
+    while (j != -1 && selected[j] != j) {
+      k = selected[j];
+      selected[j] = j;
+      j = k;
+    }
+  }
+
+  for (i = j = 0, k = -1; i < filter->used; i++) {
+    if (selected[i] != -1)
+      selected[j++] = (selected[i] + filter->used - o) % filter->used;
+  }
+
+  return j;
+}
+
 static int
 filter_get_sample(struct MedianFilter *filter, struct timeval *sample_time, double *offset, double *dispersion)
 {
-  if (filter->used == 0)
+  struct FilterSample *s, *ls;
+  int i, n, dof;
+  double x, y, d, e, var, prev_avg_var;
+
+  n = filter_select_samples(filter);
+
+  if (n < 1)
     return 0;
 
-  if (filter->used == 1) {
-    *sample_time = filter->samples[filter->index].sample_time;
-    *offset = filter->samples[filter->index].offset;
-    *dispersion = 0.0;
-  } else {
-    int i, from, to;
-    double x, x1, y, d;
+  ls = &filter->samples[filter->selected[n - 1]];
 
-    /* sort samples by offset */
-    qsort(filter->samples, filter->used, sizeof (struct FilterSample), sample_compare);
+  /* prepare data */
+  for (i = 0; i < n; i++) {
+    s = &filter->samples[filter->selected[i]];
 
-    /* average the half of the samples closest to the median */ 
-    if (filter->used > 2) {
-      from = (filter->used + 2) / 4;
-      to = filter->used - from;
-    } else {
-      from = 0;
-      to = filter->used;
-    }
-
-    for (i = from, x = y = 0.0; i < to; i++) {
-#if 0
-      LOG(LOGS_INFO, LOGF_Refclock, "refclock averaging offset %.9f [%s]",
-          filter->samples[i].offset, UTI_TimevalToString(&filter->samples[i].sample_time));
-#endif
-      UTI_DiffTimevalsToDouble(&x1, &filter->samples[i].sample_time, &filter->samples[0].sample_time);
-      x += x1;
-      y += filter->samples[i].offset;
-    }
-
-    x /= to - from;
-    y /= to - from;
-
-    for (i = from, d = 0.0; i < to; i++)
-      d += (filter->samples[i].offset - y) * (filter->samples[i].offset - y);
-
-    d = sqrt(d / (to - from));
-
-    UTI_AddDoubleToTimeval(&filter->samples[0].sample_time, x, sample_time);
-    *offset = y;
-    *dispersion = d;
+    UTI_DiffTimevalsToDouble(&filter->x_data[i], &s->sample_time, &ls->sample_time);
+    filter->y_data[i] = s->offset;
+    filter->w_data[i] = s->dispersion;
   }
 
+  /* mean offset, sample time and sample dispersion */ 
+  for (i = 0, x = y = e = 0.0; i < n; i++) {
+    x += filter->x_data[i];
+    y += filter->y_data[i];
+    e += filter->w_data[i];
+  }
+  x /= n;
+  y /= n;
+  e /= n;
+
+  e -= sqrt(filter->avg_var);
+
+  if (n >= 4) {
+    double b0, b1, s2, sb0, sb1;
+
+    /* set y axis to the mean sample time */
+    for (i = 0; i < n; i++)
+      filter->x_data[i] -= x;
+
+    /* make a linear fit and use the estimated standard deviation of intercept
+       as dispersion */
+    RGR_WeightedRegression(filter->x_data, filter->y_data, filter->w_data, n,
+        &b0, &b1, &s2, &sb0, &sb1);
+    var = s2;
+    d = sb0;
+    dof = n - 2;
+  } else if (n >= 2) {
+    for (i = 0, d = 0.0; i < n; i++)
+      d += (filter->y_data[i] - y) * (filter->y_data[i] - y);
+    var = d / (n - 1);
+    d = sqrt(var);
+    dof = n - 1;
+  } else {
+    var = filter->avg_var;
+    d = sqrt(var);
+    dof = 1;
+  }
+
+  /* avoid having zero dispersion */
+  if (var < 1e-20) {
+    var = 1e-20;
+    d = sqrt(var);
+  }
+
+  prev_avg_var = filter->avg_var;
+
+  /* update exponential moving average of the variance */
+  if (filter->avg_var_n > 50) {
+    filter->avg_var += dof / (dof + 50.0) * (var - filter->avg_var);
+  } else {
+    filter->avg_var = (filter->avg_var * filter->avg_var_n + var * dof) /
+      (dof + filter->avg_var_n);
+    if (filter->avg_var_n == 0)
+      prev_avg_var = filter->avg_var;
+    filter->avg_var_n += dof;
+  }
+
+  /* reduce noise in sourcestats weights by using the long-term average
+     instead of the estimated variance if it's not significantly lower */
+  if (var * dof / RGR_GetChi2Coef(dof) < prev_avg_var)
+    d = sqrt(filter->avg_var) * d / sqrt(var);
+
+  if (d < e)
+    d = e;
+
+  UTI_AddDoubleToTimeval(&ls->sample_time, x, sample_time);
+  *offset = y;
+  *dispersion = d;
+
+  filter_reset(filter);
+
   return 1;
 }
 
@@ -697,21 +870,29 @@ static void
 filter_slew_samples(struct MedianFilter *filter, struct timeval *when, double dfreq, double doffset)
 {
   int i;
-  double elapsed, delta_time, prev_offset;
+  double delta_time, prev_offset;
   struct timeval *sample;
 
   for (i = 0; i < filter->used; i++) {
     sample = &filter->samples[i].sample_time;
-
-    UTI_DiffTimevalsToDouble(&elapsed, when, sample);
-    delta_time = elapsed * dfreq - doffset;
-    UTI_AddDoubleToTimeval(sample, delta_time, sample);
-
+    UTI_AdjustTimeval(sample, when, sample, &delta_time, dfreq, doffset);
     prev_offset = filter->samples[i].offset;
     filter->samples[i].offset -= delta_time;
 #if 0
     LOG(LOGS_INFO, LOGF_Refclock, "i=%d old_off=%.9f new_off=%.9f",
         i, prev_offset, filter->samples[i].offset);
+#else
+    (void)prev_offset;
 #endif
   }
 }
+
+static void
+filter_add_dispersion(struct MedianFilter *filter, double dispersion)
+{
+  int i;
+
+  for (i = 0; i < filter->used; i++) {
+    filter->samples[i].dispersion += dispersion;
+  }
+}
