@@ -25,6 +25,8 @@
 
   */
 
+#include "config.h"
+
 #include "refclock.h"
 
 #if HAVE_PPSAPI
@@ -57,6 +59,8 @@ static int pps_initialise(RCL_Instance instance) {
     return 0;
   }
 
+  UTI_FdSetCloexec(fd);
+
   if (time_pps_create(fd, &handle) < 0) {
     LOG_FATAL(LOGF_Refclock, "time_pps_create() failed on %s", path);
     return 0;
@@ -93,6 +97,7 @@ static int pps_initialise(RCL_Instance instance) {
     return 0;
   }
 
+
   pps = MallocNew(struct pps_instance);
   pps->handle = handle;
   pps->last_seq = 0;
