@@ -25,6 +25,8 @@
 
   */
 
+#include "config.h"
+
 #include "refclock.h"
 #include "logging.h"
 #include "util.h"
@@ -105,6 +107,8 @@ static int sock_initialise(RCL_Instance instance)
     return 0;
   }
 
+  UTI_FdSetCloexec(sockfd);
+
   unlink(path);
   if (bind(sockfd, (struct sockaddr *)&s, sizeof (s)) < 0) {
     LOG_FATAL(LOGF_Refclock, "bind() failed");
