@@ -1,8 +1,4 @@
 /*
-  $Header: /cvs/src/chrony/rtc.c,v 1.14 2003/09/22 21:22:30 richard Exp $
-
-  =======================================================================
-
   chronyd/chronyc - Programs for keeping computer clocks accurate.
 
  **********************************************************************
@@ -27,6 +23,8 @@
 
   */
 
+#include "config.h"
+
 #include "sysincl.h"
 
 #include "rtc.h"
@@ -50,7 +48,6 @@ static struct {
   int  (*write_parameters)(void);
   int  (*get_report)(RPT_RTC_Report *report);
   int  (*trim)(void);
-  void (*cycle_logfile)(void);
 } driver =
 {
 #if defined LINUX && defined FEAT_RTC
@@ -61,8 +58,7 @@ static struct {
   RTC_Linux_StartMeasurements,
   RTC_Linux_WriteParameters,
   RTC_Linux_GetReport,
-  RTC_Linux_Trim,
-  RTC_Linux_CycleLogFile
+  RTC_Linux_Trim
 #else
   NULL,
   NULL,
@@ -71,7 +67,6 @@ static struct {
   NULL,
   NULL,
   NULL,
-  NULL,
   NULL
 #endif 
 };
@@ -89,6 +84,10 @@ RTC_Initialise(void)
   file_name = CNF_GetRtcFile();
 
   if (file_name) {
+    if (CNF_GetRTCSync()) {
+      LOG_FATAL(LOGF_Rtc, "rtcfile directive cannot be used with rtcsync");
+    }
+
     if (driver.init) {
       if ((driver.init)()) {
         ok = 1;
@@ -210,13 +209,3 @@ RTC_Trim(void)
 
 /* ================================================== */
 
-void
-RTC_CycleLogFile(void)
-{
-  if (driver_initialised) {
-    (driver.cycle_logfile)();
-  }
-}
-
-/* ================================================== */
-
