@@ -1,12 +1,9 @@
 /*
-  $Header: /cvs/src/chrony/sourcestats.c,v 1.40 2003/09/22 21:22:30 richard Exp $
-
-  =======================================================================
-
   chronyd/chronyc - Programs for keeping computer clocks accurate.
 
  **********************************************************************
  * Copyright (C) Richard P. Curnow  1997-2003
+ * Copyright (C) Miroslav Lichvar  2011
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of version 2 of the GNU General Public License as
@@ -29,6 +26,8 @@
   analysis on the samples obtained from the sources,
   to determined frequencies and error bounds. */
 
+#include "config.h"
+
 #include "sysincl.h"
 
 #include "sourcestats.h"
@@ -38,7 +37,6 @@
 #include "conf.h"
 #include "logging.h"
 #include "local.h"
-#include "mkdirpp.h"
 
 /* ================================================== */
 /* Define the maxumum number of samples that we want
@@ -49,16 +47,9 @@
    2000ppm, which would be pretty bad */
 #define WORST_CASE_FREQ_BOUND (2000.0/1.0e6)
 
-/* Day number of 1 Jan 1970 */
-#define MJD_1970 40587
-
 /* ================================================== */
-/* File to which statistics are logged, NULL if none */
-static FILE *logfile = NULL;
-static char *logfilename = NULL;
-static unsigned long logwrites = 0;
 
-#define STATISTICS_LOG "statistics.log"
+static LOG_FileID logfileid;
 
 /* ================================================== */
 /* This data structure is used to hold the history of data from the
@@ -67,19 +58,30 @@ static unsigned long logwrites = 0;
 struct SST_Stats_Record {
 
   /* Reference ID and IP address of source, used for logging to statistics log */
-  unsigned long refid;
+  uint32_t refid;
   IPAddr *ip_addr;
 
-  /* Number of samples currently stored.  sample[n_samples-1] is the
-     newest.  The samples are expected to be sorted in order, but that
-     probably doesn't matter.  */
+  /* Number of samples currently stored.  The samples are stored in circular
+     buffer. */
   int n_samples;
 
-  /* The index in the registers of the best individual sample that we
-     are holding, in terms of the minimum root distance at the present
-     time */
+  /* Number of extra samples stored in sample_times and offsets arrays that are
+     used to extend runs test */
+  int runs_samples;
+
+  /* The index of the newest sample */
+  int last_sample;
+
+  /* Flag indicating whether last regression was successful */
+  int regression_ok;
+
+  /* The best individual sample that we are holding, in terms of the minimum
+     root distance at the present time */
   int best_single_sample;
 
+  /* The index of the sample with minimum delay in peer_delays */
+  int min_delay_sample;
+
   /* This is the estimated offset (+ve => local fast) at a particular time */
   double estimated_offset;
   double estimated_offset_sd;
@@ -107,7 +109,7 @@ struct SST_Stats_Record {
 
   /* This array contains the sample epochs, in terms of the local
      clock. */
-  struct timeval sample_times[MAX_SAMPLES];
+  struct timeval sample_times[MAX_SAMPLES * REGRESS_RUNS_RATIO];
 
   /* This is an array of offsets, in seconds, corresponding to the
      sample times.  In this module, we use the convention that
@@ -115,7 +117,7 @@ struct SST_Stats_Record {
      means it is SLOW.  This is contrary to the convention in the NTP
      stuff; that part of the code is written to correspond with
      RFC1305 conventions. */
-  double offsets[MAX_SAMPLES];
+  double offsets[MAX_SAMPLES * REGRESS_RUNS_RATIO];
 
   /* This is an array of the offsets as originally measured.  Local
      clock fast of real time is indicated by positive values.  This
@@ -139,10 +141,6 @@ struct SST_Stats_Record {
      time of the measurements */
   double root_dispersions[MAX_SAMPLES];
 
-  /* This array contains the weights to be used in the regression
-     analysis for each of the samples. */
-  double weights[MAX_SAMPLES];
-
   /* This array contains the strata that were associated with the sources
      at the times the samples were generated */
   int strata[MAX_SAMPLES];
@@ -151,27 +149,17 @@ struct SST_Stats_Record {
 
 /* ================================================== */
 
+static void find_min_delay_sample(SST_Stats inst);
+static int get_buf_index(SST_Stats inst, int i);
+
+/* ================================================== */
+
 void
 SST_Initialise(void)
 {
-  char *direc;
-
-  if (CNF_GetLogStatistics()) {
-    direc = CNF_GetLogDir();
-    if (!mkdir_and_parents(direc)) {
-      LOG(LOGS_ERR, LOGF_SourceStats, "Could not create directory %s", direc);
-      logfile = NULL;
-    } else {
-      logfilename = MallocArray(char, 2 + strlen(direc) + strlen(STATISTICS_LOG));
-      strcpy(logfilename, direc);
-      strcat(logfilename, "/");
-      strcat(logfilename, STATISTICS_LOG);
-      logfile = fopen(logfilename, "a");
-      if (!logfile) {
-        LOG(LOGS_WARN, LOGF_SourceStats, "Couldn't open logfile %s for update", logfilename);
-      }
-    }
-  }
+  logfileid = CNF_GetLogStatistics() ? LOG_FileOpen("statistics",
+      "   Date (UTC) Time     IP Address    Std dev'n Est offset  Offset sd  Diff freq   Est skew  Stress  Ns  Bs  Nr")
+    : -1;
 }
 
 /* ================================================== */
@@ -179,22 +167,24 @@ SST_Initialise(void)
 void
 SST_Finalise(void)
 {
-  if (logfile) {
-    fclose(logfile);
-  }
 }
 
 /* ================================================== */
 /* This function creates a new instance of the statistics handler */
 
 SST_Stats
-SST_CreateInstance(unsigned long refid, IPAddr *addr)
+SST_CreateInstance(uint32_t refid, IPAddr *addr)
 {
   SST_Stats inst;
   inst = MallocNew(struct SST_Stats_Record);
   inst->refid = refid;
   inst->ip_addr = addr;
   inst->n_samples = 0;
+  inst->runs_samples = 0;
+  inst->last_sample = 0;
+  inst->regression_ok = 0;
+  inst->best_single_sample = 0;
+  inst->min_delay_sample = 0;
   inst->estimated_frequency = 0;
   inst->skew = 2000.0e-6;
   inst->skew_dirn = SST_Skew_Nochange;
@@ -217,45 +207,25 @@ SST_DeleteInstance(SST_Stats inst)
   return;
 }
 
-/* ================================================== */
-
-static void
-move_stats_entry(SST_Stats inst, int src, int dest)
-{
-  inst->sample_times[dest] = inst->sample_times[src];
-  inst->offsets[dest] = inst->offsets[src];
-  inst->orig_offsets[dest] = inst->orig_offsets[src];
-  inst->peer_delays[dest] = inst->peer_delays[src];
-  inst->peer_dispersions[dest] = inst->peer_dispersions[src];
-  inst->root_delays[dest] = inst->root_delays[src];
-  inst->root_dispersions[dest] = inst->root_dispersions[src];
-  inst->weights[dest] = inst->weights[src];
-  inst->strata[dest] = inst->strata[src];
-}
-
 /* ================================================== */
 /* This function is called to prune the register down when it is full.
    For now, just discard the oldest sample.  */
 
 static void
-prune_register(SST_Stats inst, int new_oldest, int *bad_points)
+prune_register(SST_Stats inst, int new_oldest)
 {
-  int i, j;
+  if (!new_oldest)
+    return;
 
-  if (!(new_oldest < inst->n_samples)) {
-    CROAK("new_oldest should be < n_samples");
-  }
-   
-  for (i=0, j=new_oldest; j<inst->n_samples; j++) {
-    if (!bad_points || !bad_points[j]) {
-      if (j != i) {
-        move_stats_entry(inst, j, i);
-      }
-      i++;
-    }
-  }
-  inst->n_samples = i;
+  assert(inst->n_samples >= new_oldest);
+  inst->n_samples -= new_oldest;
+  inst->runs_samples += new_oldest;
+  if (inst->runs_samples > inst->n_samples * (REGRESS_RUNS_RATIO - 1))
+    inst->runs_samples = inst->n_samples * (REGRESS_RUNS_RATIO - 1);
+  
+  assert(inst->n_samples + inst->runs_samples <= MAX_SAMPLES * REGRESS_RUNS_RATIO);
 
+  find_min_delay_sample(inst);
 }
 
 /* ================================================== */
@@ -267,37 +237,50 @@ SST_AccumulateSample(SST_Stats inst, struct timeval *sample_time,
                      double root_delay, double root_dispersion,
                      int stratum)
 {
-  int n;
-#if 0
-  double root_distance;
-#endif
+  int n, m;
 
   if (inst->n_samples == MAX_SAMPLES) {
-    prune_register(inst, 1, NULL);
+    prune_register(inst, 1);
   }
 
-  n = inst->n_samples;
+  n = inst->last_sample = (inst->last_sample + 1) %
+    (MAX_SAMPLES * REGRESS_RUNS_RATIO);
+  m = n % MAX_SAMPLES;
 
   inst->sample_times[n] = *sample_time;
   inst->offsets[n] = offset;
-  inst->orig_offsets[n] = offset;
-  inst->peer_delays[n] = peer_delay;
-  inst->peer_dispersions[n] = peer_dispersion;
-  inst->root_delays[n] = root_delay;
-  inst->root_dispersions[n] = root_dispersion;
-
-#if 0
-  /* The weight is worked out when we run the regression algorithm */
-  root_distance = root_dispersion + 0.5 * fabs(root_delay);
-  
-  /* For now, this is the formula for the weight functions */
-  inst->weights[n] = root_distance * root_distance;
-#endif
-
-  inst->strata[n] = stratum;
+  inst->orig_offsets[m] = offset;
+  inst->peer_delays[m] = fabs(peer_delay);
+  inst->peer_dispersions[m] = peer_dispersion;
+  inst->root_delays[m] = root_delay;
+  inst->root_dispersions[m] = root_dispersion;
+  inst->strata[m] = stratum;
  
-  ++inst->n_samples;
+  if (!inst->n_samples || inst->peer_delays[m] < inst->peer_delays[inst->min_delay_sample])
+    inst->min_delay_sample = m;
 
+  ++inst->n_samples;
+}
+
+/* ================================================== */
+/* Return index of the i-th sample in the sample_times and offset buffers,
+   i can be negative down to -runs_samples */
+
+static int
+get_runsbuf_index(SST_Stats inst, int i)
+{
+  return (unsigned int)(inst->last_sample + 2 * MAX_SAMPLES * REGRESS_RUNS_RATIO -
+      inst->n_samples + i + 1) % (MAX_SAMPLES * REGRESS_RUNS_RATIO);
+}
+
+/* ================================================== */
+/* Return index of the i-th sample in the other buffers */
+
+static int
+get_buf_index(SST_Stats inst, int i)
+{
+  return (unsigned int)(inst->last_sample + MAX_SAMPLES * REGRESS_RUNS_RATIO -
+      inst->n_samples + i + 1) % MAX_SAMPLES;
 }
 
 /* ================================================== */
@@ -311,10 +294,11 @@ convert_to_intervals(SST_Stats inst, double *times_back)
   struct timeval *newest_tv;
   int i;
 
-  newest_tv = &(inst->sample_times[inst->n_samples - 1]);
-  for (i=0; i<inst->n_samples; i++) {
+  newest_tv = &(inst->sample_times[inst->last_sample]);
+  for (i = -inst->runs_samples; i < inst->n_samples; i++) {
     /* The entries in times_back[] should end up negative */
-    UTI_DiffTimevalsToDouble(&(times_back[i]), &(inst->sample_times[i]), newest_tv);
+    UTI_DiffTimevalsToDouble(&times_back[i],
+        &inst->sample_times[get_runsbuf_index(inst, i)], newest_tv);
   }
 }
 
@@ -328,41 +312,28 @@ find_best_sample_index(SST_Stats inst, double *times_back)
 
   double root_distance, best_root_distance;
   double elapsed;
-  int i, n, best_index;
+  int i, j, best_index;
+
+  if (!inst->n_samples)
+    return;
+
+  best_index = -1;
+  best_root_distance = DBL_MAX;
+
+  for (i = 0; i < inst->n_samples; i++) {
+    j = get_buf_index(inst, i);
 
-  n = inst->n_samples - 1;
-  best_root_distance = inst->root_dispersions[n] + 0.5 * fabs(inst->root_delays[n]);
-  best_index = n;
-#if 0
-  LOG(LOGS_INFO, LOGF_SourceStats, "n=%d brd=%f", n, best_root_distance);
-#endif
-  for (i=0; i<n; i++) {
     elapsed = -times_back[i];
-#if 0
-    LOG(LOGS_INFO, LOGF_SourceStats, "n=%d i=%d latest=[%s] doing=[%s] elapsed=%f", n, i, 
-        UTI_TimevalToString(&(inst->sample_times[n])),
-        UTI_TimevalToString(&(inst->sample_times[i])),
-        elapsed);
-#endif
+    assert(elapsed >= 0.0);
 
-    /* Because the loop does not consider the most recent sample, this assertion must hold */
-    if (elapsed <= 0.0) {
-      LOG(LOGS_ERR, LOGF_SourceStats, "Elapsed<0! n=%d i=%d latest=[%s] doing=[%s] elapsed=%f",
-          n, i, 
-          UTI_TimevalToString(&(inst->sample_times[n])),
-          UTI_TimevalToString(&(inst->sample_times[i])),
-          elapsed);
-
-      elapsed = fabs(elapsed);
-    }
-
-    root_distance = inst->root_dispersions[i] + elapsed * inst->skew + 0.5 * fabs(inst->root_delays[i]);
+    root_distance = inst->root_dispersions[j] + elapsed * inst->skew + 0.5 * fabs(inst->root_delays[j]);
     if (root_distance < best_root_distance) {
       best_root_distance = root_distance;
       best_index = i;
     }
   }
-  
+
+  assert(best_index >= 0);
   inst->best_single_sample = best_index;
 
 #if 0
@@ -374,6 +345,22 @@ find_best_sample_index(SST_Stats inst, double *times_back)
 
 /* ================================================== */
 
+static void
+find_min_delay_sample(SST_Stats inst)
+{
+  int i, index;
+
+  inst->min_delay_sample = get_buf_index(inst, 0);
+
+  for (i = 1; i < inst->n_samples; i++) {
+    index = get_buf_index(inst, i);
+    if (inst->peer_delays[index] < inst->peer_delays[inst->min_delay_sample])
+      inst->min_delay_sample = index;
+  }
+}
+
+/* ================================================== */
+
 /* This defines the assumed ratio between the standard deviation of
    the samples and the peer distance as measured from the round trip
    time.  E.g. a value of 4 means that we think the standard deviation
@@ -390,53 +377,56 @@ find_best_sample_index(SST_Stats inst, double *times_back)
 void
 SST_DoNewRegression(SST_Stats inst)
 {
-  double times_back[MAX_SAMPLES];
+  double times_back[MAX_SAMPLES * REGRESS_RUNS_RATIO];
+  double offsets[MAX_SAMPLES * REGRESS_RUNS_RATIO];
   double peer_distances[MAX_SAMPLES];
+  double weights[MAX_SAMPLES];
 
-  int bad_points[MAX_SAMPLES];
   int degrees_of_freedom;
-  int best_start;
+  int best_start, times_back_start;
   double est_intercept, est_slope, est_var, est_intercept_sd, est_slope_sd;
-  int i, nruns;
-  double min_distance;
-  double sd_weight;
+  int i, j, nruns;
+  double min_distance, mean_distance;
+  double sd_weight, sd;
   double old_skew, old_freq, stress;
 
-  int regression_ok;
-
-  convert_to_intervals(inst, times_back);
+  convert_to_intervals(inst, times_back + inst->runs_samples);
 
   if (inst->n_samples > 0) {
-    for (i=0; i<inst->n_samples; i++) {
-      peer_distances[i] = 0.5 * fabs(inst->peer_delays[i]) + inst->peer_dispersions[i];
+    for (i = -inst->runs_samples; i < inst->n_samples; i++) {
+      offsets[i + inst->runs_samples] = inst->offsets[get_runsbuf_index(inst, i)];
     }
   
-    min_distance = peer_distances[0];
-    for (i=1; i<inst->n_samples; i++) {
+    for (i = 0, mean_distance = 0.0, min_distance = DBL_MAX; i < inst->n_samples; i++) {
+      j = get_buf_index(inst, i);
+      peer_distances[i] = 0.5 * inst->peer_delays[j] + inst->peer_dispersions[j];
+      mean_distance += peer_distances[i];
       if (peer_distances[i] < min_distance) {
         min_distance = peer_distances[i];
       }
     }
+    mean_distance /= inst->n_samples;
 
     /* And now, work out the weight vector */
 
-    for (i=0; i<inst->n_samples; i++) {
-      sd_weight = 1.0 + SD_TO_DIST_RATIO * (peer_distances[i] - min_distance) / min_distance;
-      inst->weights[i] = sd_weight * sd_weight;
-    } 
-  }         
+    sd = mean_distance - min_distance;
+    if (sd > min_distance || sd <= 0.0)
+      sd = min_distance;
 
-  regression_ok = RGR_FindBestRegression(times_back, inst->offsets, inst->weights,
-                                         inst->n_samples,
+    for (i=0; i<inst->n_samples; i++) {
+      sd_weight = 1.0 + SD_TO_DIST_RATIO * (peer_distances[i] - min_distance) / sd;
+      weights[i] = sd_weight * sd_weight;
+    }
+  }
+
+  inst->regression_ok = RGR_FindBestRegression(times_back + inst->runs_samples,
+                                         offsets + inst->runs_samples, weights,
+                                         inst->n_samples, inst->runs_samples,
                                          &est_intercept, &est_slope, &est_var,
                                          &est_intercept_sd, &est_slope_sd,
                                          &best_start, &nruns, &degrees_of_freedom);
 
-  /* This is a legacy of when the regression routine found outliers
-     for us.  We don't use it anymore. */
-  memset((void *) bad_points, 0, MAX_SAMPLES * sizeof(int));
-
-  if (regression_ok) {
+  if (inst->regression_ok) {
 
     old_skew = inst->skew;
     old_freq = inst->estimated_frequency;
@@ -444,7 +434,7 @@ SST_DoNewRegression(SST_Stats inst)
     inst->estimated_frequency = est_slope;
     inst->skew = est_slope_sd * RGR_GetTCoef(degrees_of_freedom);
     inst->estimated_offset = est_intercept;
-    inst->offset_time = inst->sample_times[inst->n_samples - 1];
+    inst->offset_time = inst->sample_times[inst->last_sample];
     inst->estimated_offset_sd = est_intercept_sd;
     inst->variance = est_var;
     inst->nruns = nruns;
@@ -463,17 +453,8 @@ SST_DoNewRegression(SST_Stats inst)
       }
     }
 
-    if (logfile) {
-
-      if (((logwrites++) % 32) == 0) {
-        fprintf(logfile,
-                "==============================================================================================================\n"
-                "   Date (UTC) Time     IP Address    Std dev'n Est offset  Offset sd  Diff freq   Est skew  Stress  Ns  Bs  Nr\n"
-                "==============================================================================================================\n");
-      }
-      
-            
-      fprintf(logfile, "%s %-15s %10.3e %10.3e %10.3e %10.3e %10.3e %7.1e %3d %3d %3d\n",
+    if (logfileid != -1) {
+      LOG_FileWrite(logfileid, "%s %-15s %10.3e %10.3e %10.3e %10.3e %10.3e %7.1e %3d %3d %3d",
               UTI_TimeToLogForm(inst->offset_time.tv_sec),
               inst->ip_addr ? UTI_IPToString(inst->ip_addr) : UTI_RefidToString(inst->refid),
               sqrt(inst->variance),
@@ -484,58 +465,20 @@ SST_DoNewRegression(SST_Stats inst)
               stress,
               inst->n_samples,
               best_start, nruns);
-
-      fflush(logfile);
     }
 
-    prune_register(inst, best_start, bad_points);
-
+    times_back_start = inst->runs_samples + best_start;
+    prune_register(inst, best_start);
   } else {
 #if 0
     LOG(LOGS_INFO, LOGF_SourceStats, "too few points (%d) for regression", inst->n_samples);
 #endif
     inst->estimated_frequency = 0.0;
     inst->skew = WORST_CASE_FREQ_BOUND;
+    times_back_start = 0;
   }
 
-  find_best_sample_index(inst, times_back);
-
-}
-
-/* ================================================== */
-/* This function does a simple regression on what is in the register,
-   without trying to optimise the error bounds on the frequency by
-   deleting old samples */
-
-void
-SST_DoUpdateRegression(SST_Stats inst)
-{
-  double times_back[MAX_SAMPLES];
-  double freq_error_bound;
-  double est_intercept, est_slope, est_var_base, est_intercept_sd, est_slope_sd;
-
-  convert_to_intervals(inst, times_back);
-
-  if (inst->n_samples >= 3) { /* Otherwise, we're wasting our time - we
-                                 can't do a useful linear regression
-                                 with less than 3 points */
-    
-    RGR_WeightedRegression(times_back, inst->offsets, inst->weights,
-                           inst->n_samples,
-                           &est_intercept, &est_slope, &est_var_base,
-                           &est_intercept_sd, &est_slope_sd);
-
-    freq_error_bound = est_slope_sd * RGR_GetTCoef(inst->n_samples - 2);
-
-    inst->estimated_frequency = est_slope;
-    inst->skew = freq_error_bound;
-
-  } else {
-    inst->estimated_frequency = 0.0;
-    inst->skew = WORST_CASE_FREQ_BOUND;
-  }
-
-  find_best_sample_index(inst, times_back);
+  find_best_sample_index(inst, times_back + times_back_start);
 
 }
 
@@ -549,22 +492,23 @@ SST_GetReferenceData(SST_Stats inst, struct timeval *now,
 {
 
   double elapsed;
-  int n;
+  int i, j;
 
   *frequency = inst->estimated_frequency;
   *skew = inst->skew;
 
-  n = inst->best_single_sample;
+  i = get_runsbuf_index(inst, inst->best_single_sample);
+  j = get_buf_index(inst, inst->best_single_sample);
 
-  UTI_DiffTimevalsToDouble(&elapsed, now, &(inst->sample_times[n]));
-  *root_delay = inst->root_delays[n];
-  *root_dispersion = inst->root_dispersions[n] + elapsed * inst->skew;
-  *offset = inst->offsets[n] + elapsed * inst->estimated_frequency;
-  *stratum = inst->strata[n];
+  UTI_DiffTimevalsToDouble(&elapsed, now, &inst->sample_times[i]);
+  *root_delay = inst->root_delays[j];
+  *root_dispersion = inst->root_dispersions[j] + elapsed * inst->skew;
+  *offset = inst->offsets[i] + elapsed * inst->estimated_frequency;
+  *stratum = inst->strata[j];
 
 #ifdef TRACEON
   LOG(LOGS_INFO, LOGF_SourceStats, "n=%d freq=%f skew=%f del=%f disp=%f ofs=%f str=%d",
-      n, *frequency, *skew, *root_delay, *root_dispersion, *offset, *stratum);
+      inst->n_samples, *frequency, *skew, *root_delay, *root_dispersion, *offset, *stratum);
 #endif
 
   return;
@@ -595,37 +539,45 @@ SST_GetSelectionData(SST_Stats inst, struct timeval *now,
                      int *stratum,
                      double *best_offset, double *best_root_delay,
                      double *best_root_dispersion,
-                     double *variance, int *average_ok)
+                     double *variance, int *select_ok)
 {
   double average_offset;
   double sample_elapsed;
   double elapsed;
-  int n;
+  int i, j;
+  int average_ok;
   double peer_distance;
   
-  n = inst->best_single_sample;
-  *stratum = inst->strata[n];
+  i = get_runsbuf_index(inst, inst->best_single_sample);
+  j = get_buf_index(inst, inst->best_single_sample);
+
+  *stratum = inst->strata[j];
   *variance = inst->variance;
 
-  peer_distance = inst->peer_dispersions[n] + 0.5 * fabs(inst->peer_delays[n]);
+  peer_distance = inst->peer_dispersions[j] + 0.5 * inst->peer_delays[j];
   UTI_DiffTimevalsToDouble(&elapsed, now, &(inst->offset_time));
 
-  UTI_DiffTimevalsToDouble(&sample_elapsed, now, &(inst->sample_times[n]));
-  *best_offset = inst->offsets[n] + sample_elapsed * inst->estimated_frequency;
-  *best_root_delay = inst->root_delays[n];
-  *best_root_dispersion = inst->root_dispersions[n] + sample_elapsed * inst->skew;
+  UTI_DiffTimevalsToDouble(&sample_elapsed, now, &inst->sample_times[i]);
+  *best_offset = inst->offsets[i] + sample_elapsed * inst->estimated_frequency;
+  *best_root_delay = inst->root_delays[j];
+  *best_root_dispersion = inst->root_dispersions[j] + sample_elapsed * inst->skew;
 
+  /* average_ok ignored for now */
   average_offset = inst->estimated_offset + inst->estimated_frequency * elapsed;
   if (fabs(average_offset - *best_offset) <= peer_distance) {
-    *average_ok = 1;
+    average_ok = 1;
   } else {
-    *average_ok = 0;
+    average_ok = 0;
   }
 
+  *select_ok = inst->regression_ok;
+
 #ifdef TRACEON
-  LOG(LOGS_INFO, LOGF_SourceStats, "n=%d off=%f del=%f dis=%f var=%f pdist=%f avoff=%f avok=%d",
-      n, *best_offset, *best_root_delay, *best_root_dispersion, *variance,
-      peer_distance, average_offset, *average_ok);
+  LOG(LOGS_INFO, LOGF_SourceStats, "n=%d off=%f del=%f dis=%f var=%f pdist=%f avoff=%f avok=%d selok=%d",
+      inst->n_samples, *best_offset, *best_root_delay, *best_root_dispersion, *variance,
+      peer_distance, average_offset, average_ok, *select_ok);
+#else
+  (void)average_ok;
 #endif
 
   return;
@@ -634,31 +586,30 @@ SST_GetSelectionData(SST_Stats inst, struct timeval *now,
 /* ================================================== */
 
 void
-SST_GetTrackingData(SST_Stats inst, struct timeval *now,
+SST_GetTrackingData(SST_Stats inst, struct timeval *ref_time,
                     double *average_offset, double *offset_sd,
-                    double *accrued_dispersion,
-                    double *frequency, double *skew)
+                    double *frequency, double *skew,
+                    double *root_delay, double *root_dispersion)
 {
-  int n;
-  double peer_distance;
-  double elapsed_offset, elapsed_sample;
+  int i, j;
+  double elapsed_sample;
 
-  n = inst->best_single_sample;
+  i = get_runsbuf_index(inst, inst->best_single_sample);
+  j = get_buf_index(inst, inst->best_single_sample);
 
+  *ref_time = inst->offset_time;
+  *average_offset = inst->estimated_offset;
+  *offset_sd = inst->estimated_offset_sd;
   *frequency = inst->estimated_frequency;
   *skew = inst->skew;
+  *root_delay = inst->root_delays[j];
 
-  peer_distance = inst->peer_dispersions[n] + 0.5 * fabs(inst->peer_delays[n]);
-  UTI_DiffTimevalsToDouble(&elapsed_offset, now, &(inst->offset_time));
-  *average_offset = inst->estimated_offset + inst->estimated_frequency * elapsed_offset;
-  *offset_sd = inst->estimated_offset_sd + elapsed_offset * inst->skew;
-
-  UTI_DiffTimevalsToDouble(&elapsed_sample, now, &(inst->sample_times[n]));
-  *accrued_dispersion = inst->skew * elapsed_sample;
+  UTI_DiffTimevalsToDouble(&elapsed_sample, &inst->offset_time, &inst->sample_times[i]);
+  *root_dispersion = inst->root_dispersions[j] + inst->skew * elapsed_sample;
 
 #ifdef TRACEON
-  LOG(LOGS_INFO, LOGF_SourceStats, "n=%d freq=%f (%.3fppm) skew=%f (%.3fppm) pdist=%f avoff=%f offsd=%f accrdis=%f",
-      n, *frequency, 1.0e6* *frequency, *skew, 1.0e6* *skew, peer_distance, *average_offset, *offset_sd, *accrued_dispersion);
+  LOG(LOGS_INFO, LOGF_SourceStats, "n=%d freq=%f (%.3fppm) skew=%f (%.3fppm) avoff=%f offsd=%f disp=%f",
+      inst->n_samples, *frequency, 1.0e6* *frequency, *skew, 1.0e6* *skew, *average_offset, *offset_sd, *root_dispersion);
 #endif
 
 }
@@ -668,40 +619,36 @@ SST_GetTrackingData(SST_Stats inst, struct timeval *now,
 void
 SST_SlewSamples(SST_Stats inst, struct timeval *when, double dfreq, double doffset)
 {
-  int n, i;
-  double elapsed;
+  int m, i;
   double delta_time;
   struct timeval *sample, prev;
   double prev_offset, prev_freq;
 
-  n = inst->n_samples;
+  if (!inst->n_samples)
+    return;
 
-  for (i=0; i<n; i++) {
+  for (m = -inst->runs_samples; m < inst->n_samples; m++) {
+    i = get_runsbuf_index(inst, m);
     sample = &(inst->sample_times[i]);
     prev = *sample;
-#if 0
-    UTI_AdjustTimeval(sample, when, sample, dfreq, doffset);
-    /* Can't easily use this because we need to slew offset */
-#endif
-    UTI_DiffTimevalsToDouble(&elapsed, when, sample);
-    delta_time = elapsed * dfreq - doffset;
-    UTI_AddDoubleToTimeval(sample, delta_time, sample);
+    UTI_AdjustTimeval(sample, when, sample, &delta_time, dfreq, doffset);
     prev_offset = inst->offsets[i];
     inst->offsets[i] += delta_time;
 #ifdef TRACEON
     LOG(LOGS_INFO, LOGF_SourceStats, "i=%d old_st=[%s] new_st=[%s] old_off=%f new_off=%f",
         i, UTI_TimevalToString(&prev), UTI_TimevalToString(sample),
         prev_offset, inst->offsets[i]);
+#else
+    (void)prev_offset;
 #endif
   }
 
   /* Do a half-baked update to the regression estimates */
-  UTI_DiffTimevalsToDouble(&elapsed, when, &(inst->offset_time));
   prev = inst->offset_time;
-  delta_time = elapsed * dfreq - doffset;
-  UTI_AddDoubleToTimeval(&(inst->offset_time), delta_time, &(inst->offset_time));
   prev_offset = inst->estimated_offset;
   prev_freq = inst->estimated_frequency;
+  UTI_AdjustTimeval(&(inst->offset_time), when, &(inst->offset_time),
+      &delta_time, dfreq, doffset);
   inst->estimated_offset += delta_time;
   inst->estimated_frequency -= dfreq;
 
@@ -710,6 +657,8 @@ SST_SlewSamples(SST_Stats inst, struct timeval *when, double dfreq, double doffs
       UTI_TimevalToString(&prev), UTI_TimevalToString(&(inst->offset_time)),
       prev_offset, inst->estimated_offset,
       1.0e6*prev_freq, 1.0e6*inst->estimated_frequency);
+#else
+  (void)prev; (void)prev_freq;
 #endif
 
   return;
@@ -717,6 +666,20 @@ SST_SlewSamples(SST_Stats inst, struct timeval *when, double dfreq, double doffs
 
 /* ================================================== */
 
+void 
+SST_AddDispersion(SST_Stats inst, double dispersion)
+{
+  int m, i;
+
+  for (m = 0; m < inst->n_samples; m++) {
+    i = get_buf_index(inst, m);
+    inst->root_dispersions[i] += dispersion;
+    inst->peer_dispersions[i] += dispersion;
+  }
+}
+
+/* ================================================== */
+
 double
 SST_PredictOffset(SST_Stats inst, struct timeval *when)
 {
@@ -727,7 +690,7 @@ SST_PredictOffset(SST_Stats inst, struct timeval *when)
        interval is minimal.  We can't do any useful prediction other
        than use the latest sample or zero if we don't have any samples */
     if (inst->n_samples > 0) {
-      return inst->offsets[inst->n_samples - 1];
+      return inst->offsets[inst->last_sample];
     } else {
       return 0.0;
     }
@@ -743,21 +706,48 @@ SST_PredictOffset(SST_Stats inst, struct timeval *when)
 double
 SST_MinRoundTripDelay(SST_Stats inst)
 {
-  double min_delay, delay;
-  int i;
-
-  if (inst->n_samples == 0) {
+  if (!inst->n_samples)
     return DBL_MAX;
-  } else {
-    min_delay = fabs(inst->peer_delays[0]);
-    for (i=1; i<inst->n_samples; i++) {
-      delay = fabs(inst->peer_delays[i]);
-      if (delay < min_delay) {
-        min_delay = delay;
-      }
-    }
-    return min_delay;
-  }
+  return inst->peer_delays[inst->min_delay_sample];
+}
+
+/* ================================================== */
+
+int
+SST_IsGoodSample(SST_Stats inst, double offset, double delay,
+    double max_delay_dev_ratio, double clock_error, struct timeval *when)
+{
+  double elapsed, allowed_increase, delay_increase;
+
+  if (inst->n_samples < 3)
+    return 1;
+
+  UTI_DiffTimevalsToDouble(&elapsed, when, &inst->offset_time);
+
+  /* Require that the ratio of the increase in delay from the minimum to the
+     standard deviation is less than max_delay_dev_ratio. In the allowed
+     increase in delay include also skew and clock_error. */
+    
+  allowed_increase = sqrt(inst->variance) * max_delay_dev_ratio +
+    elapsed * (inst->skew + clock_error);
+  delay_increase = (delay - SST_MinRoundTripDelay(inst)) / 2.0;
+
+  if (delay_increase < allowed_increase)
+    return 1;
+
+  offset -= inst->estimated_offset + elapsed * inst->estimated_frequency;
+
+  /* Before we decide to drop the sample, make sure the difference between
+     measured offset and predicted offset is not significantly larger than
+     the increase in delay */
+  if (fabs(offset) - delay_increase > allowed_increase)
+    return 1;
+
+#if 0
+  LOG(LOGS_INFO, LOGF_SourceStats, "bad sample: offset=%f delay=%f incr_delay=%f allowed=%f", offset, delay, allowed_increase, delay_increase);
+#endif
+
+  return 0;
 }
 
 /* ================================================== */
@@ -767,23 +757,25 @@ SST_MinRoundTripDelay(SST_Stats inst)
 void
 SST_SaveToFile(SST_Stats inst, FILE *out)
 {
-  int i;
+  int m, i, j;
 
   fprintf(out, "%d\n", inst->n_samples);
 
-  for(i=0; i<inst->n_samples; i++) {
+  for(m = 0; m < inst->n_samples; m++) {
+    i = get_runsbuf_index(inst, m);
+    j = get_buf_index(inst, m);
 
     fprintf(out, "%08lx %08lx %.6e %.6e %.6e %.6e %.6e %.6e %.6e %d\n",
             (unsigned long) inst->sample_times[i].tv_sec,
             (unsigned long) inst->sample_times[i].tv_usec,
             inst->offsets[i],
-            inst->orig_offsets[i],
-            inst->peer_delays[i],
-            inst->peer_dispersions[i],
-            inst->root_delays[i],
-            inst->root_dispersions[i],
-            inst->weights[i],
-            inst->strata[i]);
+            inst->orig_offsets[j],
+            inst->peer_delays[j],
+            inst->peer_dispersions[j],
+            inst->root_delays[j],
+            inst->root_dispersions[j],
+            1.0, /* used to be inst->weights[i] */
+            inst->strata[j]);
 
   }
 }
@@ -797,9 +789,10 @@ SST_LoadFromFile(SST_Stats inst, FILE *in)
   int i, line_number;
   char line[1024];
   unsigned long sec, usec;
+  double weight;
 
   if (fgets(line, sizeof(line), in) &&
-      (sscanf(line, "%d", &inst->n_samples) == 1)) {
+      (sscanf(line, "%u", &inst->n_samples) == 1) && inst->n_samples <= MAX_SAMPLES) {
 
     line_number = 2;
 
@@ -813,7 +806,7 @@ SST_LoadFromFile(SST_Stats inst, FILE *in)
                   &(inst->peer_dispersions[i]),
                   &(inst->root_delays[i]),
                   &(inst->root_dispersions[i]),
-                  &(inst->weights[i]),
+                  &weight, /* not used anymore */
                   &(inst->strata[i])) != 10)) {
 
         /* This is the branch taken if the read FAILED */
@@ -836,6 +829,11 @@ SST_LoadFromFile(SST_Stats inst, FILE *in)
     return 0;
   }
 
+  inst->last_sample = inst->n_samples - 1;
+  inst->runs_samples = 0;
+
+  find_min_delay_sample(inst);
+
   return 1;
 
 }
@@ -845,17 +843,18 @@ SST_LoadFromFile(SST_Stats inst, FILE *in)
 void
 SST_DoSourceReport(SST_Stats inst, RPT_SourceReport *report, struct timeval *now)
 {
-  int n;
+  int i, j;
   struct timeval ago;
 
   if (inst->n_samples > 0) {
-    n = inst->n_samples - 1;
-    report->orig_latest_meas = inst->orig_offsets[n];
-    report->latest_meas = inst->offsets[n];
-    report->latest_meas_err = 0.5*inst->root_delays[n] + inst->root_dispersions[n];
-    report->stratum = inst->strata[n];
+    i = get_runsbuf_index(inst, inst->n_samples - 1);
+    j = get_buf_index(inst, inst->n_samples - 1);
+    report->orig_latest_meas = inst->orig_offsets[j];
+    report->latest_meas = inst->offsets[i];
+    report->latest_meas_err = 0.5*inst->root_delays[j] + inst->root_dispersions[j];
+    report->stratum = inst->strata[j];
 
-    UTI_DiffTimevals(&ago, now, &inst->sample_times[n]);
+    UTI_DiffTimevals(&ago, now, &inst->sample_times[i]);
     report->latest_meas_ago = ago.tv_sec;
   } else {
     report->latest_meas_ago = 86400 * 365 * 10;
@@ -876,32 +875,43 @@ SST_Skew_Direction SST_LastSkewChange(SST_Stats inst)
 
 /* ================================================== */
 
+int
+SST_Samples(SST_Stats inst)
+{
+  return inst->n_samples;
+}
+
+/* ================================================== */
+
 void
 SST_DoSourcestatsReport(SST_Stats inst, RPT_SourcestatsReport *report, struct timeval *now)
 {
   double dspan;
   double elapsed, sample_elapsed;
-  int n, nb;
+  int li, lj, bi, bj;
 
   report->n_samples = inst->n_samples;
   report->n_runs = inst->nruns;
 
   if (inst->n_samples > 1) {
-    n = inst->n_samples - 1;
-    UTI_DiffTimevalsToDouble(&dspan, &inst->sample_times[n], &inst->sample_times[0]);
+    li = get_runsbuf_index(inst, inst->n_samples - 1);
+    lj = get_buf_index(inst, inst->n_samples - 1);
+    UTI_DiffTimevalsToDouble(&dspan, &inst->sample_times[li],
+        &inst->sample_times[get_runsbuf_index(inst, 0)]);
     report->span_seconds = (unsigned long) (dspan + 0.5);
 
     if (inst->n_samples > 3) {
       UTI_DiffTimevalsToDouble(&elapsed, now, &inst->offset_time);
-      nb = inst->best_single_sample;
-      UTI_DiffTimevalsToDouble(&sample_elapsed, now, &(inst->sample_times[nb]));
+      bi = get_runsbuf_index(inst, inst->best_single_sample);
+      bj = get_buf_index(inst, inst->best_single_sample);
+      UTI_DiffTimevalsToDouble(&sample_elapsed, now, &inst->sample_times[bi]);
       report->est_offset = inst->estimated_offset + elapsed * inst->estimated_frequency;
       report->est_offset_err = (inst->estimated_offset_sd +
                  sample_elapsed * inst->skew +
-                 (0.5*inst->root_delays[nb] + inst->root_dispersions[nb]));
+                 (0.5*inst->root_delays[bj] + inst->root_dispersions[bj]));
     } else {
-      report->est_offset = inst->offsets[n];
-      report->est_offset_err = 0.5*inst->root_delays[n] + inst->root_dispersions[n];
+      report->est_offset = inst->offsets[li];
+      report->est_offset_err = 0.5*inst->root_delays[lj] + inst->root_dispersions[lj];
     }
   } else {
     report->span_seconds = 0;
@@ -915,18 +925,3 @@ SST_DoSourcestatsReport(SST_Stats inst, RPT_SourcestatsReport *report, struct ti
 }
 
 /* ================================================== */
-
-void
-SST_CycleLogFile(void)
-{
-  if (logfile && logfilename) {
-    fclose(logfile);
-    logfile = fopen(logfilename, "a");
-    if (!logfile) {
-      LOG(LOGS_WARN, LOGF_SourceStats, "Could not reopen logfile %s", logfilename);
-    }
-    logwrites = 0;
-  }
-}
-
-/* ================================================== */
