@@ -1,8 +1,4 @@
 /*
-  $Header: /cvs/src/chrony/sys_solaris.c,v 1.19 2003/09/22 21:22:30 richard Exp $
-
-  =======================================================================
-
   chronyd/chronyc - Programs for keeping computer clocks accurate.
 
  **********************************************************************
@@ -28,6 +24,8 @@
   Driver file for Solaris operating system
   */
 
+#include "config.h"
+
 #ifdef SOLARIS
 
 #include <kvm.h>
@@ -94,24 +92,23 @@ static void
 clock_initialise(void)
 {
   struct timeval newadj, oldadj;
-  struct timezone tz;
 
   offset_register = 0.0;
   adjustment_requested = 0.0;
   current_freq = 0.0;
 
-  if (gettimeofday(&T0, &tz) < 0) {
-    CROAK("gettimeofday() failed in clock_initialise()");
+  if (gettimeofday(&T0, NULL) < 0) {
+    LOG_FATAL(LOGF_SysSolaris, "gettimeofday() failed");
   }
 
   newadj = GET_ZERO;
 
   if (adjtime(&newadj, &oldadj) < 0) {
-    CROAK("adjtime() failed in clock_initialise");
+    LOG_FATAL(LOGF_SysSolaris, "adjtime() failed");
   }
 
   if (adjtime(&newadj, &oldadj) < 0) {
-    CROAK("adjtime() failed in clock_initialise");
+    LOG_FATAL(LOGF_SysSolaris, "adjtime() failed");
   }
 
   return;
@@ -135,7 +132,6 @@ start_adjust(void)
 {
   struct timeval newadj, oldadj;
   struct timeval T1;
-  struct timezone tz;
   double elapsed, accrued_error;
   double adjust_required;
   struct timeval exact_newadj;
@@ -143,8 +139,8 @@ start_adjust(void)
   double old_adjust_remaining;
 
   /* Determine the amount of error built up since the last adjustment */
-  if (gettimeofday(&T1, &tz) < 0) {
-    CROAK("gettimeofday() failed in start_adjust");
+  if (gettimeofday(&T1, NULL) < 0) {
+    LOG_FATAL(LOGF_SysSolaris, "gettimeofday() failed");
   }
 
   UTI_DiffTimevalsToDouble(&elapsed, &T1, &T0);
@@ -164,7 +160,7 @@ start_adjust(void)
   UTI_DiffTimevalsToDouble(&rounding_error, &exact_newadj, &newadj);
 
   if (adjtime(&newadj, &oldadj) < 0) {
-    CROAK("adjtime() failed in start_adjust");
+    LOG_FATAL(LOGF_SysSolaris, "adjtime() failed");
   }
 
   UTI_TimevalToDouble(&oldadj, &old_adjust_remaining);
@@ -182,7 +178,6 @@ static void
 stop_adjust(void)
 {
   struct timeval T1;
-  struct timezone tz;
   struct timeval zeroadj, remadj;
   double adjustment_remaining, adjustment_achieved;
   double elapsed, elapsed_plus_adjust;
@@ -191,11 +186,11 @@ stop_adjust(void)
   zeroadj = GET_ZERO;
 
   if (adjtime(&zeroadj, &remadj) < 0) {
-    CROAK("adjtime() failed in stop_adjust");
+    LOG_FATAL(LOGF_SysSolaris, "adjtime() failed");
   }
 
-  if (gettimeofday(&T1, &tz) < 0) {
-    CROAK("gettimeofday() failed in stop_adjust");
+  if (gettimeofday(&T1, NULL) < 0) {
+    LOG_FATAL(LOGF_SysSolaris, "gettimeofday() failed");
   }
   
   UTI_DiffTimevalsToDouble(&elapsed, &T1, &T0);
@@ -235,11 +230,10 @@ apply_step_offset(double offset)
 {
   struct timeval old_time, new_time, rounded_new_time, T1;
   double rounding_error;
-  struct timezone tz;
   
   stop_adjust();
-  if (gettimeofday(&old_time, &tz) < 0) {
-    CROAK("gettimeofday in apply_step_offset");
+  if (gettimeofday(&old_time, NULL) < 0) {
+    LOG_FATAL(LOGF_SysSolaris, "gettimeofday() failed");
   }
 
   UTI_AddDoubleToTimeval(&old_time, -offset, &new_time);
@@ -259,8 +253,8 @@ apply_step_offset(double offset)
 
   UTI_DiffTimevalsToDouble(&rounding_error, &rounded_new_time, &new_time);
 
-  if (settimeofday(&new_time, &tz) < 0) {
-    CROAK("settimeofday in apply_step_offset");
+  if (settimeofday(&new_time, NULL) < 0) {
+    LOG_FATAL(LOGF_SysSolaris, "settimeofday() failed");
   }
 
   UTI_AddDoubleToTimeval(&T0, offset, &T1);
@@ -273,12 +267,14 @@ apply_step_offset(double offset)
 
 /* ================================================== */
 
-static void
+static double
 set_frequency(double new_freq_ppm)
 {
   stop_adjust();
   current_freq = new_freq_ppm * 1.0e-6;
   start_adjust();
+
+  return current_freq * 1.0e6;
 }
 
 /* ================================================== */
@@ -293,11 +289,13 @@ read_frequency(void)
 
 static void
 get_offset_correction(struct timeval *raw,
-                      double *corr)
+                      double *corr, double *err)
 {
   stop_adjust();
   *corr = -offset_register;
   start_adjust();
+  if (err)
+    *err = 0.0;
   return;
 }
 
@@ -390,9 +388,7 @@ set_dosynctodr(unsigned long on_off)
   kvm_t *kt;
   unsigned long read_back;
 
-  if (on_off!=1 && on_off!=0) {
-    CROAK("on_off should be 0 or 1");
-  }
+  assert(on_off == 1 || on_off == 0);
 
   kt = kvm_open(NULL, NULL, NULL, O_RDWR, NULL);
   if (!kt) {
@@ -420,9 +416,7 @@ set_dosynctodr(unsigned long on_off)
 
   kvm_close(kt);
 
-  if (read_back != on_off) {
-    CROAK("read_back should equal on_off");
-  }
+  assert(read_back == on_off);
 
 #if 0
   LOG(LOGS_INFO, LOGF_SysSolaris, "Set value of dosynctodr to %d", on_off);
