@@ -1,8 +1,4 @@
 /*
-  $Header: /cvs/src/chrony/sys_sunos.c,v 1.19 2003/09/22 21:22:30 richard Exp $
-
-  =======================================================================
-
   chronyd/chronyc - Programs for keeping computer clocks accurate.
 
  **********************************************************************
@@ -28,6 +24,8 @@
   Driver file for the SunOS 4.1.x operating system.
   */
 
+#include "config.h"
+
 #ifdef SUNOS
 
 #include <kvm.h>
@@ -84,25 +82,24 @@ static void
 clock_initialise(void)
 {
   struct timeval newadj, oldadj;
-  struct timezone tz;
 
   offset_register = 0.0;
   adjustment_requested = 0.0;
   current_freq = 0.0;
 
-  if (gettimeofday(&T0, &tz) < 0) {
-    CROAK("gettimeofday() failed in clock_initialise()");
+  if (gettimeofday(&T0, NULL) < 0) {
+    LOG_FATAL(LOGF_SysSunOS, "gettimeofday() failed");
   }
 
   newadj.tv_sec = 0;
   newadj.tv_usec = 0;
 
   if (adjtime(&newadj, &oldadj) < 0) {
-    CROAK("adjtime() failed in clock_initialise");
+    LOG_FATAL(LOGF_SysSunOS, "adjtime() failed");
   }
 
   if (adjtime(&newadj, &oldadj) < 0) {
-    CROAK("adjtime() failed in clock_initialise");
+    LOG_FATAL(LOGF_SysSunOS, "adjtime() failed");
   }
 
   return;
@@ -126,7 +123,6 @@ start_adjust(void)
 {
   struct timeval newadj, oldadj;
   struct timeval T1;
-  struct timezone tz;
   double elapsed, accrued_error;
   double adjust_required;
   struct timeval exact_newadj;
@@ -135,8 +131,8 @@ start_adjust(void)
   long remainder, multiplier;
 
   /* Determine the amount of error built up since the last adjustment */
-  if (gettimeofday(&T1, &tz) < 0) {
-    CROAK("gettimeofday() failed in start_adjust");
+  if (gettimeofday(&T1, NULL) < 0) {
+    LOG_FATAL(LOGF_SysSunOS, "gettimeofday() failed");
   }
 
   UTI_DiffTimevalsToDouble(&elapsed, &T1, &T0);
@@ -167,7 +163,7 @@ start_adjust(void)
   UTI_DiffTimevalsToDouble(&rounding_error, &newadj, &exact_newadj);
 
   if (adjtime(&newadj, &oldadj) < 0) {
-    CROAK("adjtime() failed in start_adjust");
+    LOG_FATAL(LOGF_SysSunOS, "adjtime() failed");
   }
 
   UTI_TimevalToDouble(&oldadj, &old_adjust_remaining);
@@ -185,7 +181,6 @@ static void
 stop_adjust(void)
 {
   struct timeval T1;
-  struct timezone tz;
   struct timeval zeroadj, remadj;
   double adjustment_remaining, adjustment_achieved;
   double gap;
@@ -195,11 +190,11 @@ stop_adjust(void)
   zeroadj.tv_usec = 0;
 
   if (adjtime(&zeroadj, &remadj) < 0) {
-    CROAK("adjtime() failed in stop_adjust");
+    LOG_FATAL(LOGF_SysSunOS, "adjtime() failed");
   }
 
-  if (gettimeofday(&T1, &tz) < 0) {
-    CROAK("gettimeofday() failed in stop_adjust");
+  if (gettimeofday(&T1, NULL) < 0) {
+    LOG_FATAL(LOGF_SysSunOS, "gettimeofday() failed");
   }
   
   UTI_DiffTimevalsToDouble(&elapsed, &T1, &T0);
@@ -238,17 +233,16 @@ static void
 apply_step_offset(double offset)
 {
   struct timeval old_time, new_time, T1;
-  struct timezone tz;
   
   stop_adjust();
-  if (gettimeofday(&old_time, &tz) < 0) {
-    CROAK("gettimeofday in apply_step_offset");
+  if (gettimeofday(&old_time, NULL) < 0) {
+    LOG_FATAL(LOGF_SysSunOS, "gettimeofday() failed");
   }
 
   UTI_AddDoubleToTimeval(&old_time, -offset, &new_time);
 
-  if (settimeofday(&new_time, &tz) < 0) {
-    CROAK("settimeofday in apply_step_offset");
+  if (settimeofday(&new_time, NULL) < 0) {
+    LOG_FATAL(LOGF_SysSunOS, "settimeofday() failed");
   }
 
   UTI_AddDoubleToTimeval(&T0, offset, &T1);
@@ -260,12 +254,14 @@ apply_step_offset(double offset)
 
 /* ================================================== */
 
-static void
+static double
 set_frequency(double new_freq_ppm)
 {
   stop_adjust();
   current_freq = new_freq_ppm * 1.0e-6;
   start_adjust();
+
+  return current_freq * 1.0e6;
 }
 
 /* ================================================== */
@@ -280,11 +276,13 @@ read_frequency(void)
 
 static void
 get_offset_correction(struct timeval *raw,
-                      double *corr)
+                      double *corr, double *err)
 {
   stop_adjust();
   *corr = -offset_register;
   start_adjust();
+  if (err)
+    *err = 0.0;
   return;
 }
 
@@ -339,9 +337,7 @@ setup_kernel(unsigned long on_off)
   unsigned long our_tick = 10000;
   unsigned long default_tickadj = 625;
 
-  if (on_off!=1 && on_off!=0) {
-    CROAK("on_off should be 0 or 1");
-  }
+  assert(on_off == 1 || on_off == 0);
 
   kt = kvm_open(NULL, NULL, NULL, O_RDWR, NULL);
   if (!kt) {
